"""FMP Top Losers Model."""

# pylint: disable=unused-argument

from typing import Any, Optional

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.equity_performance import (
    EquityPerformanceData,
    EquityPerformanceQueryParams,
)
from pydantic import Field, field_validator


class FMPLosersQueryParams(EquityPerformanceQueryParams):
    """FMP Losers Query.

    Source: https://site.financialmodelingprep.com/developer/docs#biggest-losers
    """


class FMPLosersData(EquityPerformanceData):
    """FMP Losers Data."""

    __alias_dict__ = {
        "percent_change": "changesPercentage",
    }

    exchange: str = Field(
        description="Stock exchange where the security is listed.",
    )

    @field_validator("percent_change", mode="before", check_fields=False)
    @classmethod
    def _normalize_percent(cls, v):
        """Normalize percent change by removing % sign and converting to float."""
        return v / 100 if v else None


class FMPLosersFetcher(Fetcher[FMPLosersQueryParams, list[FMPLosersData]]):
    """FMP Losers Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPLosersQueryParams:
        """Transform query params."""
        return FMPLosersQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPLosersQueryParams,
        credentials: Optional[dict[str, str]],
        **kwargs: Any,
    ) -> list:
        """Get the raw data from the FMP API."""
        # pylint: disable=import-outside-toplevel
        from openbb_fmp.utils.helpers import get_data_many

        api_key = credentials.get("fmp_api_key") if credentials else ""
        url = (
            f"https://financialmodelingprep.com/stable/biggest-losers?apikey={api_key}"
        )

        return await get_data_many(url)

    @staticmethod
    def transform_data(
        query: EquityPerformanceQueryParams,
        data: list[dict],
        **kwargs: Any,
    ) -> list[FMPLosersData]:
        """Transform data."""
        return [
            FMPLosersData.model_validate(d)
            for d in sorted(
                data,
                key=lambda x: x["changesPercentage"],
                reverse=query.sort == "desc",
            )
        ]



"""FMP ETF Holdings Model."""

# pylint: disable=unused-argument

from datetime import date as dateType
from typing import Any, Optional

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.nport_disclosure import (
    NportDisclosureData,
    NportDisclosureQueryParams,
)
from pydantic import Field, field_validator


class FMPNportDisclosureQueryParams(NportDisclosureQueryParams):
    """FMP ETF Holdings Query.

    Source: https://site.financialmodelingprep.com/developer/docs#holdings
    """


class FMPNportDisclosureData(NportDisclosureData):
    """FMP ETF Holdings Data."""

    __alias_dict__ = {
        "weight": "pctVal",
        "value": "valUsd",
        "asset_category": "assetCat",
        "issuer_category": "issuerCat",
        "country": "invCountry",
        "currency": "cur_cd",
        "as_of": "date",
        "is_restricted": "isRestrictedSec",
        "fair_value_level": "fairValLevel",
    }

    as_of: Optional[dateType] = Field(
        description="The acceptance datetime of the filing.",
        default=None,
    )

    @field_validator("weight", mode="before", check_fields=False)
    @classmethod
    def normalize_percent(cls, v):
        """Normalize percent values."""
        return float(v) / 100 if v else None

    @field_validator(
        "cusip",
        "isin",
        "balance",
        "name",
        "symbol",
        "value",
        mode="before",
        check_fields=False,
    )
    @classmethod
    def replace_empty(cls, v):
        """Replace empty strings and 0s with None."""
        if isinstance(v, str):
            return v if v not in ("", "0", "-", "N/A") else None
        if isinstance(v, (float, int)):
            return v if v and v not in (0.0, 0) else None
        return v if v else None


class FMPNportDisclosureFetcher(
    Fetcher[
        FMPNportDisclosureQueryParams,
        list[FMPNportDisclosureData],
    ]
):
    """FMP NPORT Disclosure Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPNportDisclosureQueryParams:
        """Transform the query."""
        return FMPNportDisclosureQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPNportDisclosureQueryParams,
        credentials: Optional[dict[str, str]],
        **kwargs: Any,
    ) -> list:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        from openbb_fmp.utils.helpers import get_data_many
        from pandas import Timestamp

        api_key = credentials.get("fmp_api_key") if credentials else ""
        url = "https://financialmodelingprep.com/stable/funds/disclosure?"

        if query.symbol.isnumeric():
            url += f"cik={query.symbol}"
        else:
            url += f"symbol={query.symbol}"

        now = Timestamp("now")

        if not query.year:
            query.year = now.year
        if not query.quarter:
            if now.quarter == 1:
                query.year -= 1
                query.quarter = 4
            else:
                query.quarter = now.quarter - 1

        url += f"&year={query.year}&quarter={query.quarter}&apikey={api_key}"

        return await get_data_many(url, **kwargs)  # type: ignore

    @staticmethod
    def transform_data(
        query: FMPNportDisclosureQueryParams,
        data: list[dict],
        **kwargs: Any,
    ) -> list[FMPNportDisclosureData]:
        """Return the transformed data."""
        results: list[FMPNportDisclosureData] = []
        for d in data:
            new_d = {k: v for k, v in d.items() if k not in ("cik", "acceptedDate")}
            results.append(FMPNportDisclosureData.model_validate(new_d))

        return sorted(results, key=lambda x: (x.weight or 0), reverse=True)



"""FMP Calendar Splits Model."""

# pylint: disable=unused-argument

from datetime import datetime, timedelta
from typing import Any, Optional

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.calendar_splits import (
    CalendarSplitsData,
    CalendarSplitsQueryParams,
)


class FMPCalendarSplitsQueryParams(CalendarSplitsQueryParams):
    """FMP Calendar Splits Query.

    Source: https://site.financialmodelingprep.com/developer/docs#splits-calendar
    """

    __alias_dict__ = {"start_date": "from", "end_date": "to"}


class FMPCalendarSplitsData(CalendarSplitsData):
    """FMP Calendar Splits Data."""


class FMPCalendarSplitsFetcher(
    Fetcher[
        FMPCalendarSplitsQueryParams,
        list[FMPCalendarSplitsData],
    ]
):
    """FMP Calendar Splits Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPCalendarSplitsQueryParams:
        """Transform the query params. Start and end dates are set to a 1 year interval."""
        return FMPCalendarSplitsQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPCalendarSplitsQueryParams,
        credentials: Optional[dict[str, str]],
        **kwargs: Any,
    ) -> list[dict]:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        from openbb_core.provider.utils.helpers import amake_requests  # noqa
        from openbb_fmp.utils.helpers import response_callback

        api_key = credentials.get("fmp_api_key") if credentials else ""

        base_url = "https://financialmodelingprep.com/stable/splits-calendar?"
        start_date = query.start_date or datetime.now().date() - timedelta(days=7)
        end_date = query.end_date or datetime.now().date() + timedelta(days=14)

        # Create 90-day chunks between start_date and end_date
        urls: list = []
        current_start = start_date

        while current_start <= end_date:
            chunk_end = min(current_start + timedelta(days=89), end_date)
            url = f"{base_url}from={current_start}&to={chunk_end}&apikey={api_key}"
            urls.append(url)
            current_start = chunk_end + timedelta(days=1)

        # Get data from all URLs
        all_data: list = await amake_requests(
            urls, response_callback=response_callback, **kwargs
        )

        return all_data

    @staticmethod
    def transform_data(
        query: FMPCalendarSplitsQueryParams, data: list[dict], **kwargs: Any
    ) -> list[FMPCalendarSplitsData]:
        """Return the transformed data."""
        return [FMPCalendarSplitsData.model_validate(d) for d in data]



"""FMP Helpers Module."""

from datetime import date
from functools import lru_cache
from typing import Any, Optional, Union

from openbb_core.app.model.abstract.error import OpenBBError
from openbb_core.provider.utils.errors import EmptyDataError, UnauthorizedError
from openbb_core.provider.utils.helpers import get_querystring


async def response_callback(response, _):
    """Use callback for make_request."""
    if response.status != 200:
        msg = await response.text()
        code = response.status
        raise UnauthorizedError(f"Unauthorized FMP request -> {code} -> {msg}")

    data = await response.json()

    if isinstance(data, dict):
        error_message = data.get("Error Message", data.get("error"))

        if error_message is not None:
            conditions = (
                "upgrade" in error_message.lower()
                or "exclusive endpoint" in error_message.lower()
                or "special endpoint" in error_message.lower()
                or "premium query parameter" in error_message.lower()
                or "subscription" in error_message.lower()
                or "unauthorized" in error_message.lower()
                or "premium" in error_message.lower()
            )

            if conditions:
                raise UnauthorizedError(f"Unauthorized FMP request -> {error_message}")

            raise OpenBBError(
                f"FMP Error Message -> Status code: {response.status} -> {error_message}"
            )

    return data


async def get_data(url: str, **kwargs: Any) -> Union[list, dict]:
    """Get data from FMP endpoint."""
    # pylint: disable=import-outside-toplevel
    from openbb_core.provider.utils.helpers import amake_request

    return await amake_request(url, response_callback=response_callback, **kwargs)


async def get_data_urls(urls: list[str], **kwargs: Any) -> Union[list, dict]:
    """Get data from FMP for several urls."""
    # pylint: disable=import-outside-toplevel
    from openbb_core.provider.utils.helpers import amake_requests

    return await amake_requests(urls, response_callback=response_callback, **kwargs)


def create_url(
    version: int,
    endpoint: str,
    api_key: Optional[str],
    query: Optional[Any] = None,
    exclude: Optional[list[str]] = None,
) -> str:
    """Return a URL for the FMP API.

    Parameters
    ----------
    version: int
        The version of the API to use.
    endpoint: str
        The endpoint to use.
    api_key: str
        The API key to use.
    query: Optional[BaseModel]
        The dictionary to be turned into a querystring.
    exclude: list[str]
        The keys to be excluded from the querystring.

    Returns
    -------
    str
        The querystring.
    """
    # pylint: disable=import-outside-toplevel
    from pydantic import BaseModel

    the_dict = {}
    if query:
        the_dict = query.model_dump() if isinstance(query, BaseModel) else query
    query_string = get_querystring(the_dict, exclude or [])
    base_url = f"https://financialmodelingprep.com/api/v{version}/"
    return f"{base_url}{endpoint}?{query_string}&apikey={api_key}"


async def get_data_many(
    url: str, sub_dict: Optional[str] = None, **kwargs: Any
) -> list[dict]:
    """Get data from FMP endpoint and convert to list of schemas.

    Parameters
    ----------
    url: str
        The URL to get the data from.
    sub_dict: Optional[str]
        The sub-dictionary to use.

    Returns
    -------
    list[dict]
        Dictionary of data.
    """
    data = await get_data(url, **kwargs)

    if sub_dict and isinstance(data, dict):
        data = data.get(sub_dict, [])
    if isinstance(data, dict):
        raise ValueError("Expected list of dicts, got dict")
    if len(data) == 0:
        raise EmptyDataError()

    return data


async def get_data_one(url: str, **kwargs: Any) -> dict:
    """Get data from FMP endpoint and convert to schema."""
    data = await get_data(url, **kwargs)
    if isinstance(data, list):
        if len(data) == 0:
            raise ValueError("Expected dict, got empty list")

        try:
            data = {i: data[i] for i in range(len(data))} if len(data) > 1 else data[0]
        except TypeError as e:
            raise ValueError("Expected dict, got list of dicts") from e

    return data


def most_recent_quarter(base: Optional[date] = None) -> date:
    """Get the most recent quarter date."""
    if base is None:
        base = date.today()
    base = min(base, date.today())  # This prevents dates from being in the future
    exacts = [(3, 31), (6, 30), (9, 30), (12, 31)]
    for exact in exacts:
        if base.month == exact[0] and base.day == exact[1]:
            return base
    if base.month < 4:
        return date(base.year - 1, 12, 31)
    if base.month < 7:
        return date(base.year, 3, 31)
    if base.month < 10:
        return date(base.year, 6, 30)
    return date(base.year, 9, 30)


def get_interval(value: str) -> str:
    """Get the intervals for the FMP API."""
    intervals = {
        "m": "min",
        "h": "hour",
        "d": "day",
    }

    return f"{value[:-1]}{intervals[value[-1]]}"


async def get_historical_ohlc(query, credentials, **kwargs: Any) -> list[dict]:
    """Return the raw data from the FMP endpoint."""
    # pylint: disable=import-outside-toplevel
    import asyncio  # noqa
    from openbb_core.provider.utils.helpers import (
        amake_request,
    )
    from warnings import warn

    api_key = credentials.get("fmp_api_key") if credentials else ""

    base_url = "https://financialmodelingprep.com/stable/"

    if hasattr(query, "adjustment") and query.adjustment == "unadjusted":
        base_url += "historical-price-eod/non-split-adjusted?"
    elif hasattr(query, "adjustment") and query.adjustment == "splits_and_dividends":
        base_url += "historical-price-eod/dividend-adjusted?"
    elif query.interval == "1d":
        base_url += "historical-price-eod/full?"
    elif query.interval == "1m":
        base_url += "historical-chart/1min?"
    elif query.interval == "5m":
        base_url += "historical-chart/5min?"
    elif query.interval in ["60m", "1h"]:
        query.interval = "60m"
        base_url += "historical-chart/1hour?"

    query_str = get_querystring(
        query.model_dump(), ["symbol", "adjustment", "interval"]
    )
    symbols = query.symbol.split(",")

    results: list = []
    messages: list = []

    async def get_one(symbol):
        """Get data for one symbol."""
        url = f"{base_url}symbol={symbol}&{query_str}&apikey={api_key}"
        data: list = []
        response = await amake_request(
            url, response_callback=response_callback, **kwargs
        )

        if isinstance(response, dict) and response.get("Error Message"):
            message = (
                f"Error fetching data for {symbol}: {response.get('Error Message', '')}"
            )
            warn(message)
            messages.append(message)

        if isinstance(response, list) and len(response) > 0:
            data = response

        elif isinstance(response, dict) and response.get("historical"):
            data = response.get("historical", [])

        if not data:
            message = f"No data found for {symbol}."
            warn(message)
            messages.append(message)

        elif data:
            for d in data:
                d["symbol"] = symbol
                results.append(d)

    await asyncio.gather(*[get_one(symbol) for symbol in symbols])

    if not results:
        raise EmptyDataError(
            f"{str(','.join(messages)).replace(',',' ') if messages else 'No data found'}"
        )

    return results


@lru_cache(maxsize=1)
def get_available_transcript_symbols(api_key) -> list:
    """Return the available symbols for earnings call transcripts."""
    # pylint: disable=import-outside-toplevel
    from openbb_core.provider.utils.helpers import make_request

    url = f"https://financialmodelingprep.com/stable/earnings-transcript-list?apikey={api_key}"

    data = make_request(url)

    data.raise_for_status()

    return data.json()


@lru_cache(maxsize=64)
def get_transcript_dates_for_symbol(symbol: str, api_key: str) -> list:
    """Return the available dates for a given symbol's earnings call transcripts."""
    # pylint: disable=import-outside-toplevel
    from openbb_core.provider.utils.helpers import make_request

    url = f"https://financialmodelingprep.com/stable/earning-call-transcript-dates?symbol={symbol}&apikey={api_key}"

    data = make_request(url)

    data.raise_for_status()

    return data.json()



"""FMP Government Trades Model."""

# pylint: disable=unused-argument,too-many-locals

from typing import Any, Literal, Optional

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.government_trades import (
    GovernmentTradesData,
    GovernmentTradesQueryParams,
)
from pydantic import ConfigDict, Field, model_validator


class FMPGovernmentTradesQueryParams(GovernmentTradesQueryParams):
    """Government Trades Query Parameters.

    Source: https://site.financialmodelingprep.com/developer/docs#house-trading
    """

    __json_schema_extra__ = {"symbol": {"multiple_items_allowed": True}}


class FMPGovernmentTradesData(GovernmentTradesData):
    """Government Trades Data Model."""

    model_config = ConfigDict(extra="ignore")

    __alias_dict__ = {
        "symbol": "ticker",
        "transaction_date": "transactionDate",
        "representative": "office",
        "url": "link",
        "transaction_type": "type",
        "date": "disclosureDate",
    }
    chamber: Literal["House", "Senate"] = Field(
        description="Government Chamber - House or Senate."
    )
    owner: Optional[str] = Field(
        default=None, description="Ownership status (e.g., Spouse, Joint)."
    )
    asset_type: Optional[str] = Field(
        default=None, description="Type of asset involved in the transaction."
    )
    asset_description: Optional[str] = Field(
        default=None, description="Description of the asset."
    )
    transaction_type: Optional[str] = Field(
        default=None, description="Type of transaction (e.g., Sale, Purchase)."
    )
    amount: Optional[str] = Field(default=None, description="Transaction amount range.")
    comment: Optional[str] = Field(
        default=None, description="Additional comments on the transaction."
    )
    url: Optional[str] = Field(
        default=None, description="Link to the transaction document."
    )

    @model_validator(mode="before")
    @classmethod
    def _fill_missing(cls, values):
        """Fill missing information that can be identified."""
        description = values.get("assetDescription", "").lower()
        if not values.get("owner"):
            values["owner"] = "Self"
        if (values.get("ticker") or values.get("symbol")) and not values.get(
            "assetType"
        ):
            values["asset_type"] = "ETF" if "etf" in description else "Stock"
        elif (
            not values.get("ticker")
            and not values.get("symbol")
            and not values.get("assetType")
        ):
            values["asset_type"] = (
                "Treasury"
                if "treasury" in description or "bill" in description
                else (
                    "Bond"
                    if "%" in description
                    or "due" in description
                    or "pct" in description
                    else (
                        "Fund"
                        if "fund" in description
                        else ("ETF" if "etf" in description else None)
                    )
                )
            )
        return values


class FMPGovernmentTradesFetcher(
    Fetcher[
        FMPGovernmentTradesQueryParams,
        list[FMPGovernmentTradesData],
    ]
):
    """FMP Government Trades Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPGovernmentTradesQueryParams:
        """Transform the query params."""
        return FMPGovernmentTradesQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPGovernmentTradesQueryParams,
        credentials: Optional[dict[str, str]] = None,
        **kwargs: Any,
    ) -> list[dict]:
        """Return the raw data from the Government Trades endpoint."""
        # pylint: disable=import-outside-toplevel
        import asyncio  # noqa
        from openbb_core.app.model.abstract.error import OpenBBError
        from openbb_core.provider.utils.errors import EmptyDataError
        from openbb_core.provider.utils.helpers import amake_request
        from openbb_fmp.utils.helpers import response_callback
        from warnings import warn

        symbols: list = []

        if query.symbol:
            symbols = query.symbol.split(",")

        results: list[dict] = []
        chamber_url_dict = {
            "house": ["house-trades"],
            "senate": ["senate-trades"],
            "all": ["house-trades", "senate-trades"],
        }
        api_key = credentials.get("fmp_api_key") if credentials else ""
        keys_to_remove = {
            "district",
            "capitalGainsOver200USD",
            "disclosureYear",
            "firstName",
            "lastName",
        }
        keys_to_rename = {"dateRecieved": "date", "disclosureDate": "date"}

        async def get_one(url):
            """Get data for one URL."""
            result = await amake_request(
                url, response_callback=response_callback, **kwargs
            )
            processed_list: list = []

            for entry in result:
                new_entry = {
                    keys_to_rename.get(k, k): v
                    for k, v in entry.items()
                    if k not in keys_to_remove
                }
                new_entry["chamber"] = "Senate" if "senate-trades" in url else "House"
                processed_list.append(new_entry)

            if not processed_list or len(processed_list) == 0:
                warn(f"No data found for {url.replace(api_key, 'API_KEY')}")

            if processed_list:
                results.extend(processed_list)

        urls_list: list = []
        base_url = "https://financialmodelingprep.com/stable/"
        limit = query.limit if query.limit else 1000
        try:
            if symbols:
                for symbol in symbols:
                    query.symbol = symbol
                    url = [
                        f"{base_url}{i}?symbol={symbol}&apikey={api_key}"
                        for i in chamber_url_dict[query.chamber]
                    ]
                    urls_list.extend(url)
                await asyncio.gather(*[get_one(url) for url in urls_list])
            else:
                page = 0
                seen = set()
                unique_results: list = []

                while len(unique_results) < limit:
                    all_urls = []
                    for i in chamber_url_dict[query.chamber]:
                        chamber = i.split("-")[0]
                        page_url = f"{base_url}{i.replace('trades', 'latest')}?page={page}&limit=250&apikey={api_key}"
                        all_urls.append((page_url, chamber.title()))

                    async def fetch_page(url_info):
                        url, chamber_name = url_info
                        try:
                            result = await amake_request(
                                url, response_callback=response_callback, **kwargs
                            )
                            if result:
                                for d in result:
                                    d["chamber"] = chamber_name
                                return result
                        except Exception:
                            return []
                        return []

                    page_results = await asyncio.gather(
                        *[fetch_page(url_info) for url_info in all_urls],
                        return_exceptions=True,
                    )

                    new_data_found = False
                    for page_result in page_results:
                        if isinstance(page_result, list) and page_result:
                            new_data_found = True
                            for d in page_result:
                                fs = frozenset(d.items())
                                if fs not in seen:
                                    seen.add(fs)
                                    unique_results.append(d)
                                    if len(unique_results) >= limit:
                                        break
                            if len(unique_results) >= limit:
                                break

                    if not new_data_found:
                        break  # No more data from any source
                    page += 1

                results = unique_results

            if not results:
                raise EmptyDataError("No data returned for the given symbol.")

            return results
        except OpenBBError as e:
            raise e from e

    @staticmethod
    def transform_data(
        query: FMPGovernmentTradesQueryParams, data: list[dict], **kwargs: Any
    ) -> list[FMPGovernmentTradesData]:
        """Return the transformed data."""
        return sorted(
            [
                FMPGovernmentTradesData(
                    **{k: v for k, v in d.items() if v and v != "--"}
                )
                for d in data
            ],
            key=lambda x: x.date,
            reverse=True,
        )[: query.limit or len(data)]



"""FMP Equity Ownership Model."""

# pylint: disable=unused-argument

from typing import Any, Optional

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.price_target import (
    PriceTargetData,
    PriceTargetQueryParams,
)
from openbb_core.provider.utils.errors import EmptyDataError
from pydantic import ConfigDict, Field


class FMPPriceTargetQueryParams(PriceTargetQueryParams):
    """FMP Price Target Query.

    Source: https://site.financialmodelingprep.com/developer/docs#analyst
    """

    __json_schema_extra__ = {"symbol": {"multiple_items_allowed": True}}


class FMPPriceTargetData(PriceTargetData):
    """FMP Price Target Data."""

    model_config = ConfigDict(extra="ignore")

    __alias_dict__ = {
        "analyst_firm": "analystCompany",
        "rating_current": "newGrade",
        "rating_previous": "previousGrade",
        "news_title": "newsTitle",
        "news_url": "newsURL",
    }

    news_title: Optional[str] = Field(
        default=None, description="News title of the price target."
    )
    news_url: Optional[str] = Field(
        default=None, description="News URL of the price target."
    )


class FMPPriceTargetFetcher(
    Fetcher[
        FMPPriceTargetQueryParams,
        list[FMPPriceTargetData],
    ]
):
    """FMP Price Target Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPPriceTargetQueryParams:
        """Transform the query params."""
        return FMPPriceTargetQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPPriceTargetQueryParams,
        credentials: Optional[dict[str, str]],
        **kwargs: Any,
    ) -> list[dict]:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        import asyncio  # noqa
        import math
        from warnings import warn
        from openbb_fmp.utils.helpers import get_data_urls

        api_key = credentials.get("fmp_api_key") if credentials else ""
        base_url = "https://financialmodelingprep.com/stable/price-target-news?"
        symbols = query.symbol.split(",")  # type: ignore
        limit = query.limit if query.limit else 100
        pages = math.ceil(limit / 100)
        results: list[dict] = []

        async def get_one(symbol):
            """Get data for one symbol."""
            urls = [
                f"{base_url}symbol={symbol}&page={page}&limit={limit}&apikey={api_key}"
                for page in range(pages)
            ]
            result = await get_data_urls(urls, **kwargs)

            if not result or len(result) == 0:
                warn(f"Symbol Error: No data found for {symbol}")

            if result:
                results.extend(result)

        await asyncio.gather(*[get_one(symbol) for symbol in symbols])

        if not results:
            raise EmptyDataError("No data returned for the given symbols.")

        return sorted(
            results, key=lambda x: (x["publishedDate"], x["symbol"]), reverse=True
        )

    @staticmethod
    def transform_data(
        query: FMPPriceTargetQueryParams, data: list[dict], **kwargs: Any
    ) -> list[FMPPriceTargetData]:
        """Return the transformed data."""
        return [FMPPriceTargetData.model_validate(item) for item in data]



"""FMP Index Constituents Model."""

# pylint: disable=unused-argument

from datetime import (
    date as dateType,
    datetime,
)
from typing import Any, Literal, Optional, Union

from dateutil import parser
from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.index_constituents import (
    IndexConstituentsData,
    IndexConstituentsQueryParams,
)
from openbb_core.provider.utils.descriptions import DATA_DESCRIPTIONS
from pydantic import Field, field_validator


class FMPIndexConstituentsQueryParams(IndexConstituentsQueryParams):
    """FMP Index Constituents Query.

    Source: https://site.financialmodelingprep.com/developer/docs#sp-500
    """

    symbol: Literal["dowjones", "sp500", "nasdaq"] = Field(
        default="dowjones",
    )
    historical: bool = Field(
        default=False,
        description="Flag to retrieve historical removals and additions.",
    )


class FMPIndexConstituentsData(IndexConstituentsData):
    """FMP Index Constituents Data."""

    __alias_dict__ = {
        "headquarter": "headQuarter",
        "date_added": "dateFirstAdded",
        "industry": "subSector",
        "name": "addedSecurity",
        "removed_symbol": "removedTicker",
        "removed_name": "removedSecurity",
    }

    sector: Optional[str] = Field(
        default=None,
        description="Sector classification for the constituent company in the index.",
    )
    industry: Optional[str] = Field(
        default=None,
        description="Industry classification for the constituent company in the index.",
    )
    headquarter: Optional[str] = Field(
        default=None,
        description="Location of the company's headquarters.",
    )
    date_added: Optional[Union[dateType, str]] = Field(
        default=None, description="Date the constituent company was added to the index."
    )
    cik: Optional[str] = Field(
        description=DATA_DESCRIPTIONS.get("cik", ""),
        default=None,
        coerce_numbers_to_str=True,
    )
    founded: Optional[Union[dateType, str]] = Field(
        default=None,
        description="When the company was founded.",
    )
    removed_symbol: Optional[str] = Field(
        default=None,
        description="Symbol of the company removed from the index.",
    )
    removed_name: Optional[str] = Field(
        default=None,
        description="Name of the company removed from the index.",
    )
    reason: Optional[str] = Field(
        default=None,
        description="Reason for the removal from the index.",
    )
    date: Optional[dateType] = Field(
        default=None,
        description="Date of the historical constituent data.",
    )

    @field_validator("date_added", "founded", "date", mode="before", check_fields=False)
    @classmethod
    def date_first_added_validate(cls, v):
        """Return the date_first_added date as a datetime object for valid cases."""
        if not v:
            return None

        try:
            # First try ISO format for performance
            return datetime.fromisoformat(str(v)).date()
        except (ValueError, TypeError):
            try:
                # Fall back to dateutil parser for flexible parsing
                return parser.parse(str(v)).date()
            except Exception:
                # Return as string if all parsing fails
                return str(v)

    @field_validator(
        "removed_symbol", "removed_name", "reason", mode="before", check_fields=False
    )
    @classmethod
    def _clean_empty_strings(cls, v):  # pylint: disable=E0213
        """Return the removed fields as strings."""
        if not v or v in ("''", "", "None"):
            return None
        return v


class FMPIndexConstituentsFetcher(
    Fetcher[
        FMPIndexConstituentsQueryParams,
        list[FMPIndexConstituentsData],
    ]
):
    """FMP Index Constituents Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPIndexConstituentsQueryParams:
        """Transform the query params."""
        return FMPIndexConstituentsQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPIndexConstituentsQueryParams,
        credentials: Optional[dict[str, str]],
        **kwargs: Any,
    ) -> list:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        from openbb_fmp.utils.helpers import get_data_many

        api_key = credentials.get("fmp_api_key") if credentials else ""

        base_url = "https://financialmodelingprep.com/stable"
        url = f"{base_url}/{'historical-' if query.historical else ''}{query.symbol}-constituent/?apikey={api_key}"

        return await get_data_many(url, **kwargs)

    @staticmethod
    def transform_data(
        query: FMPIndexConstituentsQueryParams, data: list[dict], **kwargs: Any
    ) -> list[FMPIndexConstituentsData]:
        """Transform the raw data into a list of FMPIndexConstituentsData."""
        return [FMPIndexConstituentsData.model_validate(d) for d in data]



"""FMP Company Filings Model."""

# pylint: disable=unused-argument

from datetime import (
    date as dateType,
    timedelta,
)
from typing import Any, Optional

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.company_filings import (
    CompanyFilingsData,
    CompanyFilingsQueryParams,
)
from openbb_core.provider.utils.descriptions import (
    DATA_DESCRIPTIONS,
)
from openbb_core.provider.utils.errors import EmptyDataError
from pydantic import Field


class FMPCompanyFilingsQueryParams(CompanyFilingsQueryParams):
    """FMP Company Filings Query.

    Source: https://site.financialmodelingprep.com/developer/docs/sec-filings-api/
    """

    cik: Optional[str] = Field(
        default=None, description="CIK number to look up. Overrides symbol."
    )
    start_date: Optional[dateType] = Field(
        default=None,
        description="Start date for filtering filings. Default is one year ago.",
    )
    end_date: Optional[dateType] = Field(
        default=None, description="End date for filtering filings."
    )
    limit: int = Field(
        default=1000,
        le=1000,
        gt=0,
        description="Number of results to return. Max results is 1000.",
    )
    page: int = Field(
        default=0,
        description="Page number for paginated results. Max page is 100.",
        le=100,
        gt=0,
    )


class FMPCompanyFilingsData(CompanyFilingsData):
    """FMP Company Filings Data."""

    __alias_dict__ = {
        "accepted_date": "acceptedDate",
        "report_type": "formType",
        "filing_url": "link",
        "report_url": "finalLink",
    }
    filing_url: Optional[str] = Field(
        default=None, description="URL to the filing page."
    )
    symbol: Optional[str] = Field(
        default=None, description=DATA_DESCRIPTIONS.get("symbol", "")
    )
    cik: Optional[str] = Field(
        default=None, description=DATA_DESCRIPTIONS.get("cik", "")
    )
    accepted_date: Optional[dateType] = Field(
        default=None, description="Accepted date of the filing."
    )


class FMPCompanyFilingsFetcher(
    Fetcher[
        FMPCompanyFilingsQueryParams,
        list[FMPCompanyFilingsData],
    ]
):
    """FMP Company Filings Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPCompanyFilingsQueryParams:
        """Transform the query params."""
        return FMPCompanyFilingsQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPCompanyFilingsQueryParams,
        credentials: Optional[dict[str, str]],
        **kwargs: Any,
    ) -> list[dict]:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        from openbb_fmp.utils.helpers import get_data_many

        api_key = credentials.get("fmp_api_key") if credentials else ""

        base_url = "https://financialmodelingprep.com/stable/sec-filings-search"
        url: str = ""

        if query.symbol and not query.cik:
            url = base_url + f"/symbol?symbol={query.symbol}"
        elif query.cik:
            url = base_url + f"/cik?cik={query.cik}"

        if not url:
            raise ValueError("Either symbol or cik must be provided.")

        start_date = (
            query.start_date
            if query.start_date
            else dateType.today() - timedelta(days=360)
        )
        url += f"&from={start_date}"
        end_date = query.end_date if query.end_date else dateType.today()
        url += f"&to={end_date}"
        url += f"&page={query.page}&limit={query.limit}&apikey={api_key}"

        return await get_data_many(url, **kwargs)

    @staticmethod
    def transform_data(
        query: FMPCompanyFilingsQueryParams, data: list[dict], **kwargs: Any
    ) -> list[FMPCompanyFilingsData]:
        """Return the transformed data."""
        if not data:
            raise EmptyDataError(
                f"No data found for the given query -> {query.model_dump()}"
            )
        return [FMPCompanyFilingsData.model_validate(d) for d in data]



"""FMP Equity Valuation Multiples Model."""

# pylint: disable=unused-argument

from datetime import datetime
from typing import Any, Literal, Optional, Union

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.key_metrics import (
    KeyMetricsData,
    KeyMetricsQueryParams,
)
from openbb_core.provider.utils.errors import EmptyDataError
from openbb_fmp.utils.definitions import FinancialPeriods
from pydantic import Field


class FMPKeyMetricsQueryParams(KeyMetricsQueryParams):
    """FMP Equity Valuation Multiples Query.

    Source: https://site.financialmodelingprep.com/developer/docs#key-metrics
    """

    __json_schema_extra__ = {"symbol": {"multiple_items_allowed": True}}

    ttm: Literal["include", "exclude", "only"] = Field(
        default="only",
        description="Specify whether to include, exclude, or only show TTM (Trailing Twelve Months) data."
        + " The default is 'only'.",
    )
    period: FinancialPeriods = Field(
        default="annual",
        description="Specify the fiscal period for the data. Ignored when TTM is set to 'only'.",
    )
    limit: Optional[int] = Field(
        default=None,
        description="Only applicable when TTM is not set to 'only'."
        + " Defines the number of most recent reporting periods to return."
        + " The default is 5.",
    )


class FMPKeyMetricsData(KeyMetricsData):
    """FMP Equity Valuation Multiples Data."""

    __alias_dict__ = {
        "period_ending": "date",
        "fiscal_period": "period",
        "currency": "reportedCurrency",
        "enterprise_value": "enterpriseValueTTM",
        "ev_to_sales": "evToSalesTTM",
        "ev_to_operating_cash_flow": "evToOperatingCashFlowTTM",
        "ev_to_free_cash_flow": "evToFreeCashFlowTTM",
        "ev_to_ebitda": "evToEBITDATTM",
        "net_debt_to_ebitda": "netDebtToEBITDATTM",
        "current_ratio": "currentRatioTTM",
        "income_quality": "incomeQualityTTM",
        "graham_number": "grahamNumberTTM",
        "graham_net_net": "grahamNetNetTTM",
        "tax_burden": "taxBurdenTTM",
        "interest_burden": "interestBurdenTTM",
        "working_capital": "workingCapitalTTM",
        "invested_capital": "investedCapitalTTM",
        "return_on_assets": "returnOnAssetsTTM",
        "operating_return_on_assets": "operatingReturnOnAssetsTTM",
        "return_on_tangible_assets": "returnOnTangibleAssetsTTM",
        "return_on_equity": "returnOnEquityTTM",
        "return_on_invested_capital": "returnOnInvestedCapitalTTM",
        "return_on_capital_employed": "returnOnCapitalEmployedTTM",
        "earnings_yield": "earningsYieldTTM",
        "free_cash_flow_yield": "freeCashFlowYieldTTM",
        "capex_to_operating_cash_flow": "capexToOperatingCashFlowTTM",
        "capex_to_depreciation": "capexToDepreciationTTM",
        "capex_to_revenue": "capexToRevenueTTM",
        "sales_general_and_administrative_to_revenue": "salesGeneralAndAdministrativeToRevenueTTM",
        "research_and_developement_to_revenue": "researchAndDevelopementToRevenueTTM",
        "stock_based_compensation_to_revenue": "stockBasedCompensationToRevenueTTM",
        "intangibles_to_total_assets": "intangiblesToTotalAssetsTTM",
        "average_receivables": "averageReceivablesTTM",
        "average_payables": "averagePayablesTTM",
        "average_inventory": "averageInventoryTTM",
        "days_of_sales_outstanding": "daysOfSalesOutstandingTTM",
        "days_of_payables_outstanding": "daysOfPayablesOutstandingTTM",
        "days_of_inventory_outstanding": "daysOfInventoryOutstandingTTM",
        "operating_cycle": "operatingCycleTTM",
        "cash_conversion_cycle": "cashConversionCycleTTM",
        "free_cash_flow_to_equity": "freeCashFlowToEquityTTM",
        "free_cash_flow_to_firm": "freeCashFlowToFirmTTM",
        "tangible_asset_value": "tangibleAssetValueTTM",
        "net_current_asset_value": "netCurrentAssetValueTTM",
    }

    enterprise_value: Optional[Union[int, float]] = Field(
        default=None, description="Enterprise Value."
    )
    ev_to_sales: Optional[float] = Field(
        default=None, description="Enterprise Value to Sales ratio.", title="EV/Sales"
    )
    ev_to_operating_cash_flow: Optional[float] = Field(
        default=None,
        description="Enterprise Value to Operating Cash Flow ratio.",
        title="EV/Operating Cash Flow",
    )
    ev_to_free_cash_flow: Optional[float] = Field(
        default=None,
        description="Enterprise Value to Free Cash Flow ratio.",
        title="EV/Free Cash Flow",
    )
    ev_to_ebitda: Optional[float] = Field(
        default=None,
        description="Enterprise Value to EBITDA ratio.",
        title="EV/EBITDA",
        alias="evToEBITDA",
    )
    net_debt_to_ebitda: Optional[float] = Field(
        default=None,
        description="Net Debt to EBITDA ratio.",
        title="Net Debt/EBITDA",
        alias="netDebtToEBITDA",
    )
    current_ratio: Optional[float] = Field(default=None, description="Current Ratio.")
    income_quality: Optional[float] = Field(default=None, description="Income Quality.")
    graham_number: Optional[float] = Field(default=None, description="Graham Number.")
    graham_net_net: Optional[float] = Field(default=None, description="Graham Net Net.")
    tax_burden: Optional[float] = Field(default=None, description="Tax Burden.")
    interest_burden: Optional[float] = Field(
        default=None, description="Interest Burden."
    )
    working_capital: Optional[Union[int, float]] = Field(
        default=None, description="Working Capital."
    )
    invested_capital: Optional[Union[int, float]] = Field(
        default=None, description="Invested Capital."
    )
    return_on_assets: Optional[float] = Field(
        default=None,
        description="Return on Assets.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    operating_return_on_assets: Optional[float] = Field(
        default=None,
        description="Operating Return on Assets.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    return_on_tangible_assets: Optional[float] = Field(
        default=None,
        description="Return on Tangible Assets.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    return_on_equity: Optional[float] = Field(
        default=None,
        description="Return on Equity.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    return_on_invested_capital: Optional[float] = Field(
        default=None,
        description="Return on Invested Capital.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    return_on_capital_employed: Optional[float] = Field(
        default=None,
        description="Return on Capital Employed.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    earnings_yield: Optional[float] = Field(
        default=None,
        description="Earnings Yield.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    free_cash_flow_yield: Optional[float] = Field(
        default=None,
        description="Free Cash Flow Yield.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    capex_to_operating_cash_flow: Optional[float] = Field(
        default=None, description="Capex to Operating Cash Flow."
    )
    capex_to_depreciation: Optional[float] = Field(
        default=None, description="Capex to Depreciation."
    )
    capex_to_revenue: Optional[float] = Field(
        default=None, description="Capex to Revenue."
    )
    sales_general_and_administrative_to_revenue: Optional[float] = Field(
        default=None,
        description="Sales, General and Administrative to Revenue.",
        title="SG&A to Revenue",
    )
    research_and_developement_to_revenue: Optional[float] = Field(
        default=None,
        description="Research and Development to Revenue.",
        title="R&D to Revenue",
    )
    stock_based_compensation_to_revenue: Optional[float] = Field(
        default=None, description="Stock Based Compensation to Revenue."
    )
    intangibles_to_total_assets: Optional[float] = Field(
        default=None, description="Intangibles to Total Assets."
    )
    average_receivables: Optional[Union[int, float]] = Field(
        default=None, description="Average Receivables."
    )
    average_payables: Optional[Union[int, float]] = Field(
        default=None, description="Average Payables."
    )
    average_inventory: Optional[Union[int, float]] = Field(
        default=None, description="Average Inventory."
    )
    days_of_sales_outstanding: Optional[float] = Field(
        default=None, description="Days of Sales Outstanding."
    )
    days_of_payables_outstanding: Optional[float] = Field(
        default=None, description="Days of Payables Outstanding."
    )
    days_of_inventory_outstanding: Optional[float] = Field(
        default=None, description="Days of Inventory Outstanding."
    )
    operating_cycle: Optional[float] = Field(
        default=None, description="Operating Cycle."
    )
    cash_conversion_cycle: Optional[float] = Field(
        default=None, description="Cash Conversion Cycle."
    )
    free_cash_flow_to_equity: Optional[float] = Field(
        default=None, description="Free Cash Flow to Equity."
    )
    free_cash_flow_to_firm: Optional[float] = Field(
        default=None, description="Free Cash Flow to Firm."
    )
    tangible_asset_value: Optional[Union[int, float]] = Field(
        default=None, description="Tangible Asset Value."
    )
    net_current_asset_value: Optional[Union[int, float]] = Field(
        default=None, description="Net Current Asset Value."
    )


class FMPKeyMetricsFetcher(
    Fetcher[
        FMPKeyMetricsQueryParams,
        list[FMPKeyMetricsData],
    ]
):
    """FMP Key Metrics Fetcher."""

    @staticmethod
    def transform_query(
        params: dict[str, Any],
    ) -> FMPKeyMetricsQueryParams:
        """Transform the query params."""
        return FMPKeyMetricsQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPKeyMetricsQueryParams,
        credentials: Optional[dict[str, str]],
        **kwargs: Any,
    ) -> list[dict]:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        import asyncio  # noqa
        import warnings
        from openbb_fmp.utils.helpers import get_data_many

        api_key = credentials.get("fmp_api_key") if credentials else ""
        symbols = query.symbol.split(",")
        results: list = []
        base_url: str = "https://financialmodelingprep.com/stable/key-metrics"
        limit = query.limit if query.limit and query.ttm != "only" else 1

        async def get_one(symbol):
            """Get data for one symbol."""
            ttm = f"{base_url}-ttm?symbol={symbol}&apikey={api_key}"
            metrics = f"{base_url}?symbol={symbol}&period={query.period}&limit={limit}&apikey={api_key}"
            result: list = []
            ttm_data = await get_data_many(ttm, **kwargs)
            metrics_data = await get_data_many(metrics, **kwargs)
            currency = None

            if metrics_data:
                if query.ttm != "only":
                    result.extend(metrics_data)
                currency = metrics_data[0].get("reportedCurrency")

            if ttm_data and query.ttm != "exclude":
                ttm_result = ttm_data[0]
                ttm_result["date"] = datetime.today().date().isoformat()
                ttm_result["fiscal_period"] = "TTM"
                ttm_result["fiscal_year"] = datetime.today().year
                if currency:
                    ttm_result["reportedCurrency"] = currency
                result.insert(0, ttm_result)

            if not result:
                warnings.warn(f"Symbol Error: No data found for {symbol}.")

            if result:
                results.extend(result)

        tasks = [get_one(symbol) for symbol in symbols]

        await asyncio.gather(*tasks)

        if not results:
            raise EmptyDataError("No data found for given symbols.")

        return results

    @staticmethod
    def transform_data(
        query: FMPKeyMetricsQueryParams,
        data: list[dict],
        **kwargs: Any,
    ) -> list[FMPKeyMetricsData]:
        """Return the transformed data."""
        return [
            FMPKeyMetricsData.model_validate(d)
            for d in sorted(data, key=lambda x: x["date"], reverse=True)
        ]



"""FMP Market Snapshots Model."""

# pylint: disable=unused-argument

from datetime import (
    date as dateType,
    datetime,
)
from typing import Any, Optional, Union

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.market_snapshots import (
    MarketSnapshotsData,
    MarketSnapshotsQueryParams,
)
from openbb_fmp.utils.definitions import MARKETS
from pydantic import Field, field_validator


class FMPMarketSnapshotsQueryParams(MarketSnapshotsQueryParams):
    """FMP Market Snapshots Query.

    Source: https://site.financialmodelingprep.com/developer/docs#exchange-prices-quote
    """

    __json_schema_extra__ = {
        "market": {
            "x-widget_config": {
                "options": [{"label": m.upper(), "value": m} for m in MARKETS.__args__]
            }
        }
    }

    market: MARKETS = Field(
        description="The market to fetch data for.", default="nasdaq"
    )


class FMPMarketSnapshotsData(MarketSnapshotsData):
    """FMP Market Snapshots Data."""

    __alias_dict__ = {
        "high": "dayHigh",
        "low": "dayLow",
        "prev_close": "previousClose",
        "change_percent": "changePercentage",
        "close": "price",
        "last_price_timestamp": "timestamp",
        "ma50": "priceAvg50",
        "ma200": "priceAvg200",
        "year_high": "yearHigh",
        "year_low": "yearLow",
        "market_cap": "marketCap",
    }
    ma50: Optional[float] = Field(
        description="The 50-day moving average.", default=None
    )
    ma200: Optional[float] = Field(
        description="The 200-day moving average.", default=None
    )
    year_high: Optional[float] = Field(description="The 52-week high.", default=None)
    year_low: Optional[float] = Field(description="The 52-week low.", default=None)
    market_cap: Optional[Union[int, float]] = Field(
        description="Market cap of the stock.", default=None
    )
    last_price_timestamp: Optional[Union[datetime, dateType]] = Field(
        description="The timestamp of the last price.", default=None
    )

    @field_validator("change_percent", mode="before", check_fields=False)
    @classmethod
    def _normalize_percent(cls, v):
        """Normalize the percent."""
        return float(v) / 100 if v else 0

    @field_validator("name", mode="before", check_fields=False)
    @classmethod
    def _empty_strings(cls, v):
        """Clear empty strings."""
        return v if v and v not in (" ", "''") else None


class FMPMarketSnapshotsFetcher(
    Fetcher[
        FMPMarketSnapshotsQueryParams,
        list[FMPMarketSnapshotsData],
    ]
):
    """FMP Market Snapshots Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPMarketSnapshotsQueryParams:
        """Transform the query params."""
        return FMPMarketSnapshotsQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPMarketSnapshotsQueryParams,
        credentials: Optional[dict[str, str]],
        **kwargs: Any,
    ) -> list:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        from openbb_fmp.utils.helpers import get_data_many

        api_key = credentials.get("fmp_api_key") if credentials else ""
        base_url = "https://financialmodelingprep.com/stable/batch-"
        market = query.market.upper()

        if market == "ETF":
            url = f"{base_url}etf-quotes?short=false&apikey={api_key}"
        elif market == "MUTUAL_FUND":
            url = f"{base_url}mutualfund-quotes?short=false&apikey={api_key}"
        elif market == "FOREX":
            url = f"{base_url}forex-quotes?short=false&apikey={api_key}"
        elif market == "CRYPTO":
            url = f"{base_url}crypto-quotes?short=false&apikey={api_key}"
        elif market == "INDEX":
            url = f"{base_url}index-quotes?short=false&apikey={api_key}"
        elif market == "COMMODITY":
            url = f"{base_url}commodity-quotes?short=false&apikey={api_key}"
        else:
            url = f"{base_url}exchange-quote?exchange={market}&short=false&apikey={api_key}"

        return await get_data_many(url, **kwargs)

    @staticmethod
    def transform_data(
        query: FMPMarketSnapshotsQueryParams, data: list, **kwargs: Any
    ) -> list[FMPMarketSnapshotsData]:
        """Return the transformed data."""
        # pylint: disable=import-outside-toplevel
        import pandas as pd
        from openbb_core.provider.utils.errors import EmptyDataError

        df = pd.DataFrame(data)

        if df.empty:
            raise EmptyDataError("No data was returned")

        # We need to clean up the response because there is lots of very old data included.
        # Purge to most recent day only
        df.timestamp = pd.to_datetime(df.timestamp, unit="s", utc=True).dt.tz_localize(
            None
        )
        max_date = df.timestamp.max().date()
        df = df[df.timestamp.dt.date == max_date]

        return [
            FMPMarketSnapshotsData.model_validate(d)
            for d in df.sort_values(by="timestamp", ascending=False).to_dict(
                orient="records"
            )
        ]



"""FMP Historical Splits Model."""

# pylint: disable=unused-argument

from datetime import datetime
from typing import Any, Optional

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.historical_splits import (
    HistoricalSplitsData,
    HistoricalSplitsQueryParams,
)
from openbb_fmp.utils.helpers import get_data_many
from pydantic import field_validator


class FMPHistoricalSplitsQueryParams(HistoricalSplitsQueryParams):
    """FMP Historical Splits Query.

    Source: https://site.financialmodelingprep.com/developer/docs#splits-company
    """


class FMPHistoricalSplitsData(HistoricalSplitsData):
    """FMP Historical Splits Data."""

    @field_validator("date", mode="before", check_fields=False)
    @classmethod
    def date_validate(cls, v: str):  # pylint: disable=E0213
        """Return the date as a datetime object."""
        return datetime.strptime(v, "%Y-%m-%d") if v else None


class FMPHistoricalSplitsFetcher(
    Fetcher[
        FMPHistoricalSplitsQueryParams,
        list[FMPHistoricalSplitsData],
    ]
):
    """FMP Historical Splits Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPHistoricalSplitsQueryParams:
        """Transform the query params."""
        return FMPHistoricalSplitsQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPHistoricalSplitsQueryParams,
        credentials: Optional[dict[str, str]],
        **kwargs: Any,
    ) -> list[dict]:
        """Return the raw data from the FMP endpoint."""
        api_key = credentials.get("fmp_api_key") if credentials else ""

        url = f"https://financialmodelingprep.com/stable/splits?symbol={query.symbol}&apikey={api_key}"

        return await get_data_many(url, "historical", **kwargs)

    @staticmethod
    def transform_data(
        query: FMPHistoricalSplitsQueryParams, data: list[dict], **kwargs: Any
    ) -> list[FMPHistoricalSplitsData]:
        """Return the transformed data."""
        return [FMPHistoricalSplitsData.model_validate(d) for d in data]



"""FMP Equity Peers Model."""

# pylint: disable=unused-argument

from typing import Any, Optional

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.equity_peers import (
    EquityPeersData,
    EquityPeersQueryParams,
)
from pydantic import Field


class FMPEquityPeersQueryParams(EquityPeersQueryParams):
    """FMP Equity Peers Query.

    Source: https://site.financialmodelingprep.com/developer/docs#peers
    """


class FMPEquityPeersData(EquityPeersData):
    """FMP Equity Peers Data."""

    __alias_dict__ = {"name": "companyName", "market_cap": "mktCap"}

    name: str = Field(
        description="The name of the company.",
    )
    price: Optional[float] = Field(
        default=None,
        description="The current stock price of the company.",
    )
    market_cap: Optional[int] = Field(
        default=None,
        description="The market capitalization of the company.",
    )


class FMPEquityPeersFetcher(
    Fetcher[
        FMPEquityPeersQueryParams,
        list[FMPEquityPeersData],
    ]
):
    """FMP Equity Peers Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPEquityPeersQueryParams:
        """Transform the query params."""
        return FMPEquityPeersQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPEquityPeersQueryParams,
        credentials: Optional[dict[str, str]],
        **kwargs: Any,
    ) -> list:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        from openbb_fmp.utils.helpers import get_data_many

        api_key = credentials.get("fmp_api_key") if credentials else ""
        url = (
            "https://financialmodelingprep.com/stable/stock-peers?"
            f"symbol={query.symbol}&apikey={api_key}"
        )

        return await get_data_many(url, **kwargs)

    @staticmethod
    def transform_data(
        query: FMPEquityPeersQueryParams, data: list, **kwargs: Any
    ) -> list[FMPEquityPeersData]:
        """Return the transformed data."""
        return [
            FMPEquityPeersData.model_validate(d)
            for d in sorted(data, key=lambda x: x["mktCap"], reverse=True)
        ]



"""FMP Currency Historical Price Model."""

# pylint: disable=unused-argument

from datetime import datetime
from typing import Any, Literal, Optional

from dateutil.relativedelta import relativedelta
from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.currency_historical import (
    CurrencyHistoricalData,
    CurrencyHistoricalQueryParams,
)
from openbb_core.provider.utils.descriptions import (
    QUERY_DESCRIPTIONS,
)
from pydantic import Field, field_validator


class FMPCurrencyHistoricalQueryParams(CurrencyHistoricalQueryParams):
    """FMP Currency Historical Price Query.

    Source: https://site.financialmodelingprep.com/developer/docs#forex-historical-price-eod-full
    """

    __alias_dict__ = {"start_date": "from", "end_date": "to"}
    __json_schema_extra__ = {
        "symbol": {"multiple_items_allowed": True},
    }

    interval: Literal["1m", "5m", "1h", "1d"] = Field(
        default="1d", description=QUERY_DESCRIPTIONS.get("interval", "")
    )


class FMPCurrencyHistoricalData(CurrencyHistoricalData):
    """FMP Currency Historical Price Data."""

    change: Optional[float] = Field(
        default=None,
        description="Change in the price from the previous close.",
    )
    change_percent: Optional[float] = Field(
        default=None,
        description="Percent change in the price from the previous close.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )

    @field_validator("change_percent", mode="before", check_fields=False)
    @classmethod
    def _normalize_percent(cls, v):
        """Normalize percent."""
        return v / 100 if v else None


class FMPCurrencyHistoricalFetcher(
    Fetcher[
        FMPCurrencyHistoricalQueryParams,
        list[FMPCurrencyHistoricalData],
    ]
):
    """FMP Currency Historical Price Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPCurrencyHistoricalQueryParams:
        """Transform the query params. Start and end dates are set to a 1 year interval."""
        transformed_params = params

        now = datetime.now().date()
        if params.get("start_date") is None:
            transformed_params["start_date"] = now - relativedelta(years=1)

        if params.get("end_date") is None:
            transformed_params["end_date"] = now

        return FMPCurrencyHistoricalQueryParams(**transformed_params)

    @staticmethod
    async def aextract_data(
        query: FMPCurrencyHistoricalQueryParams,
        credentials: Optional[dict[str, str]],
        **kwargs: Any,
    ) -> list[dict]:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        from openbb_fmp.utils.helpers import get_historical_ohlc

        return await get_historical_ohlc(query, credentials, **kwargs)

    @staticmethod
    def transform_data(
        query: FMPCurrencyHistoricalQueryParams, data: list[dict], **kwargs: Any
    ) -> list[FMPCurrencyHistoricalData]:
        """Return the transformed data."""
        return [
            FMPCurrencyHistoricalData.model_validate(d)
            for d in sorted(
                data,
                key=lambda x: (
                    (x["date"], x["symbol"])
                    if len(query.symbol.split(",")) > 1
                    else x["date"]
                ),
                reverse=False,
            )
        ]



"""FMP Balance Sheet Model."""

# pylint: disable=unused-argument

from datetime import (
    date as dateType,
    datetime,
)
from typing import Any, Optional

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.balance_sheet import (
    BalanceSheetData,
    BalanceSheetQueryParams,
)
from openbb_core.provider.utils.descriptions import QUERY_DESCRIPTIONS
from openbb_fmp.utils.definitions import FinancialStatementPeriods
from pydantic import Field


class FMPBalanceSheetQueryParams(BalanceSheetQueryParams):
    """FMP Balance Sheet Query.

    Source: https://site.financialmodelingprep.com/developer/docs#balance-sheet-statement
    """

    period: FinancialStatementPeriods = Field(
        default="annual",
        description=QUERY_DESCRIPTIONS.get("period", ""),
    )


class FMPBalanceSheetData(BalanceSheetData):
    """FMP Balance Sheet Data."""

    __alias_dict__ = {
        "period_ending": "date",
        "fiscal_period": "period",
        "fiscal_year": "calendarYear",
        "filing_date": "fillingDate",
        "accepted_date": "acceptedDate",
        "reported_currency": "reportedCurrency",
        "cash_and_cash_equivalents": "cashAndCashEquivalents",
        "short_term_investments": "shortTermInvestments",
        "cash_and_short_term_investments": "cashAndShortTermInvestments",
        "net_receivables": "netReceivables",
        "inventory": "inventories",
        "other_current_assets": "otherCurrentAssets",
        "total_current_assets": "totalCurrentAssets",
        "plant_property_equipment_net": "propertyPlantEquipmentNet",
        "goodwill": "goodwill",
        "prepaid_expenses": "prepaids",
        "intangible_assets": "intangibleAssets",
        "goodwill_and_intangible_assets": "goodwillAndIntangibleAssets",
        "long_term_investments": "longTermInvestments",
        "tax_assets": "taxAssets",
        "other_non_current_assets": "otherNonCurrentAssets",
        "non_current_assets": "totalNonCurrentAssets",
        "other_assets": "otherAssets",
        "total_assets": "totalAssets",
        "accounts_payable": "accountPayables",
        "short_term_debt": "shortTermDebt",
        "tax_payables": "taxPayables",
        "current_deferred_revenue": "deferredRevenue",
        "other_current_liabilities": "otherCurrentLiabilities",
        "total_current_liabilities": "totalCurrentLiabilities",
        "long_term_debt": "longTermDebt",
        "deferred_revenue_non_current": "deferredRevenueNonCurrent",
        "deferred_tax_liabilities_non_current": "deferredTaxLiabilitiesNonCurrent",
        "other_non_current_liabilities": "otherNonCurrentLiabilities",
        "total_non_current_liabilities": "totalNonCurrentLiabilities",
        "other_liabilities": "otherLiabilities",
        "capital_lease_obligations": "capitalLeaseObligations",
        "total_liabilities": "totalLiabilities",
        "preferred_stock": "preferredStock",
        "common_stock": "commonStock",
        "retained_earnings": "retainedEarnings",
        "accumulated_other_comprehensive_income": "accumulatedOtherComprehensiveIncomeLoss",
        "other_shareholders_equity": "otherStockholdersEquity",
        "other_total_shareholders_equity": "otherTotalStockholdersEquity",
        "total_common_equity": "totalStockholdersEquity",
        "total_equity_non_controlling_interests": "totalEquity",
        "total_liabilities_and_shareholders_equity": "totalLiabilitiesAndStockholdersEquity",
        "minority_interest": "minorityInterest",
        "total_liabilities_and_total_equity": "totalLiabilitiesAndTotalEquity",
        "total_investments": "totalInvestments",
        "total_debt": "totalDebt",
        "net_debt": "netDebt",
    }

    filing_date: Optional[dateType] = Field(
        default=None,
        description="The date when the filing was made.",
    )
    accepted_date: Optional[datetime] = Field(
        default=None,
        description="The date and time when the filing was accepted.",
    )
    cik: Optional[str] = Field(
        default=None,
        description="The Central Index Key (CIK) assigned by the SEC, if applicable.",
    )
    symbol: Optional[str] = Field(
        default=None,
        description="The stock ticker symbol.",
    )
    reported_currency: Optional[str] = Field(
        default=None,
        description="The currency in which the balance sheet was reported.",
    )
    cash_and_cash_equivalents: Optional[int] = Field(
        default=None,
        description="Cash and cash equivalents.",
    )
    short_term_investments: Optional[int] = Field(
        default=None,
        description="Short term investments.",
    )
    cash_and_short_term_investments: Optional[int] = Field(
        default=None,
        description="Cash and short term investments.",
    )
    accounts_receivables: Optional[int] = Field(
        default=None,
        description="Accounts receivables.",
    )
    other_receivables: Optional[int] = Field(
        default=None,
        description="Other receivables.",
    )
    net_receivables: Optional[int] = Field(
        default=None,
        description="Net receivables.",
    )
    inventory: Optional[int] = Field(
        default=None,
        description="Inventory.",
    )
    other_current_assets: Optional[int] = Field(
        default=None,
        description="Other current assets.",
    )
    total_current_assets: Optional[int] = Field(
        default=None,
        description="Total current assets.",
    )
    plant_property_equipment_net: Optional[int] = Field(
        default=None,
        description="Plant property equipment net.",
    )
    goodwill: Optional[int] = Field(
        default=None,
        description="Goodwill.",
    )
    intangible_assets: Optional[int] = Field(
        default=None,
        description="Intangible assets.",
    )
    goodwill_and_intangible_assets: Optional[int] = Field(
        default=None,
        description="Goodwill and intangible assets.",
    )
    long_term_investments: Optional[int] = Field(
        default=None,
        description="Long term investments.",
    )
    tax_assets: Optional[int] = Field(
        default=None,
        description="Tax assets.",
    )
    other_non_current_assets: Optional[int] = Field(
        default=None,
        description="Other non current assets.",
    )
    non_current_assets: Optional[int] = Field(
        default=None,
        description="Total non current assets.",
    )
    other_assets: Optional[int] = Field(
        default=None,
        description="Other assets.",
    )
    total_assets: Optional[int] = Field(
        default=None,
        description="Total assets.",
    )
    accounts_payable: Optional[int] = Field(
        default=None,
        description="Accounts payable.",
    )
    prepaid_expenses: Optional[int] = Field(
        default=None,
        description="Prepaid expenses.",
    )
    accrued_expenses: Optional[int] = Field(
        default=None,
        description="Accrued expenses.",
    )
    short_term_debt: Optional[int] = Field(
        default=None,
        description="Short term debt.",
    )
    tax_payables: Optional[int] = Field(
        default=None,
        description="Tax payables.",
    )
    current_deferred_revenue: Optional[int] = Field(
        default=None,
        description="Current deferred revenue.",
    )
    other_current_liabilities: Optional[int] = Field(
        default=None,
        description="Other current liabilities.",
    )
    other_payables: Optional[int] = Field(
        default=None,
        description="Other payables.",
    )
    total_current_liabilities: Optional[int] = Field(
        default=None,
        description="Total current liabilities.",
    )
    total_payables: Optional[int] = Field(
        default=None,
        description="Total payables.",
    )
    long_term_debt: Optional[int] = Field(
        default=None,
        description="Long term debt.",
    )
    deferred_revenue_non_current: Optional[int] = Field(
        default=None,
        description="Non current deferred revenue.",
    )
    deferred_tax_liabilities_non_current: Optional[int] = Field(
        default=None,
        description="Deferred tax liabilities non current.",
    )
    other_non_current_liabilities: Optional[int] = Field(
        default=None,
        description="Other non current liabilities.",
    )
    total_non_current_liabilities: Optional[int] = Field(
        default=None,
        description="Total non current liabilities.",
    )

    capital_lease_obligations_current: Optional[int] = Field(
        default=None,
        description="Current capital lease obligations.",
    )
    capital_lease_obligations_non_current: Optional[int] = Field(
        default=None,
        description="Non current capital lease obligations.",
    )
    capital_lease_obligations: Optional[int] = Field(
        default=None,
        description="Capital lease obligations.",
    )
    other_liabilities: Optional[int] = Field(
        default=None,
        description="Other liabilities.",
    )
    total_liabilities: Optional[int] = Field(
        default=None,
        description="Total liabilities.",
    )
    preferred_stock: Optional[int] = Field(
        default=None,
        description="Preferred stock.",
    )
    common_stock: Optional[int] = Field(
        default=None,
        description="Common stock.",
    )
    treasury_stock: Optional[int] = Field(
        default=None,
        description="Treasury stock.",
    )
    retained_earnings: Optional[int] = Field(
        default=None,
        description="Retained earnings.",
    )
    additional_paid_in_capital: Optional[int] = Field(
        default=None,
        description="Additional paid in capital.",
    )
    accumulated_other_comprehensive_income: Optional[int] = Field(
        default=None,
        description="Accumulated other comprehensive income (loss).",
    )
    other_shareholders_equity: Optional[int] = Field(
        default=None,
        description="Other shareholders equity.",
    )
    other_total_shareholders_equity: Optional[int] = Field(
        default=None,
        description="Other total shareholders equity.",
    )
    total_common_equity: Optional[int] = Field(
        default=None,
        description="Total common equity.",
    )
    total_equity_non_controlling_interests: Optional[int] = Field(
        default=None,
        description="Total equity non controlling interests.",
    )
    total_liabilities_and_shareholders_equity: Optional[int] = Field(
        default=None,
        description="Total liabilities and shareholders equity.",
    )
    minority_interest: Optional[int] = Field(
        default=None,
        description="Minority interest.",
    )
    total_liabilities_and_total_equity: Optional[int] = Field(
        default=None,
        description="Total liabilities and total equity.",
    )
    total_investments: Optional[int] = Field(
        default=None,
        description="Total investments.",
    )
    total_debt: Optional[int] = Field(
        default=None,
        description="Total debt.",
    )
    net_debt: Optional[int] = Field(
        default=None,
        description="Net debt.",
    )


class FMPBalanceSheetFetcher(
    Fetcher[
        FMPBalanceSheetQueryParams,
        list[FMPBalanceSheetData],
    ]
):
    """Transform the query, extract and transform the data from the FMP endpoints."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPBalanceSheetQueryParams:
        """Transform the query params."""
        return FMPBalanceSheetQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPBalanceSheetQueryParams,
        credentials: Optional[dict[str, str]],
        **kwargs: Any,
    ) -> list[dict]:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        from openbb_fmp.utils.helpers import get_data_many

        api_key = credentials.get("fmp_api_key") if credentials else ""
        base_url = "https://financialmodelingprep.com/stable/balance-sheet-statement"

        if query.period == "ttm":
            base_url += "-ttm"

        url = (
            base_url
            + f"?symbol={query.symbol}{'&period=' + query.period if query.period != 'ttm' else ''}"
            + f"&limit={query.limit if query.limit else 5}"
            + f"&apikey={api_key}"
        )

        return await get_data_many(url, **kwargs)

    @staticmethod
    def transform_data(
        query: FMPBalanceSheetQueryParams, data: list[dict], **kwargs: Any
    ) -> list[FMPBalanceSheetData]:
        """Return the transformed data."""
        return [FMPBalanceSheetData.model_validate(d) for d in data]



"""FMP Company News Model."""

from typing import Any, Optional

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.company_news import (
    CompanyNewsData,
    CompanyNewsQueryParams,
)
from openbb_core.provider.utils.errors import EmptyDataError
from pydantic import Field, field_validator


class FMPCompanyNewsQueryParams(CompanyNewsQueryParams):
    """FMP Company News Query.

    Source: https://site.financialmodelingprep.com/developer/docs/stock-news-api/
    """

    __json_schema_extra__ = {"symbol": {"multiple_items_allowed": True}}

    page: int = Field(
        default=0,
        le=100,
        ge=0,
        description="Page number of the results. Use in combination with limit.",
    )
    press_release: Optional[bool] = Field(
        default=None,
        description="When true, will return only press releases for the given symbol(s).",
    )

    @field_validator("symbol", mode="before", check_fields=False)
    @classmethod
    def _symbol_mandatory(cls, v):
        """Symbol mandatory validator."""
        if not v:
            raise ValueError("Required field missing -> symbol")
        return v


class FMPCompanyNewsData(CompanyNewsData):
    """FMP Company News Data."""

    __alias_dict__ = {
        "symbols": "symbol",
        "date": "publishedDate",
        "author": "publisher",
        "images": "image",
        "source": "site",
        "excerpt": "text",
    }

    source: str = Field(description="Name of the news site.")


class FMPCompanyNewsFetcher(
    Fetcher[
        FMPCompanyNewsQueryParams,
        list[FMPCompanyNewsData],
    ]
):
    """Transform the query, extract and transform the data from the FMP endpoints."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPCompanyNewsQueryParams:
        """Transform the query params."""
        return FMPCompanyNewsQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPCompanyNewsQueryParams,
        credentials: Optional[dict[str, str]],
        **kwargs: Any,
    ) -> list[dict]:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        from openbb_fmp.utils.helpers import get_data_many

        api_key = credentials.get("fmp_api_key") if credentials else ""
        limit = query.limit if query.limit else 250
        page = query.page if query.page else 0
        base_url = "https://financialmodelingprep.com/stable/news/"
        symbols = query.symbol

        if query.press_release:
            base_url += "press-releases?"
        else:
            base_url += "stock?"

        url = base_url + f"symbols={symbols}"

        if query.start_date:
            url += f"&from={query.start_date}"

        if query.end_date:
            url += f"&to={query.end_date}"

        url += f"&limit={limit}&page={page}&apikey={api_key}"

        response = await get_data_many(url, **kwargs)

        if not response:
            raise EmptyDataError()

        return sorted(response, key=lambda x: x["publishedDate"], reverse=True)

    # pylint: disable=unused-argument
    @staticmethod
    def transform_data(
        query: FMPCompanyNewsQueryParams, data: list[dict], **kwargs: Any
    ) -> list[FMPCompanyNewsData]:
        """Return the transformed data."""
        return [FMPCompanyNewsData.model_validate(d) for d in data]



"""FMP Key Executives Model."""

# pylint: disable=unused-argument

from typing import Any, Optional

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.key_executives import (
    KeyExecutivesData,
    KeyExecutivesQueryParams,
)
from pydantic import ConfigDict


class FMPKeyExecutivesQueryParams(KeyExecutivesQueryParams):
    """FMP Key Executives Query.

    Source: https://site.financialmodelingprep.com/developer/docs#company-executives
    """


class FMPKeyExecutivesData(KeyExecutivesData):
    """FMP Key Executives Data."""

    model_config = ConfigDict(extra="ignore")


class FMPKeyExecutivesFetcher(
    Fetcher[
        FMPKeyExecutivesQueryParams,
        list[FMPKeyExecutivesData],
    ]
):
    """FMP Key Executives Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPKeyExecutivesQueryParams:
        """Transform the query params."""
        return FMPKeyExecutivesQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPKeyExecutivesQueryParams,
        credentials: Optional[dict[str, str]],
        **kwargs: Any,
    ) -> list[dict]:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        from openbb_fmp.utils.helpers import get_data_many

        api_key = credentials.get("fmp_api_key") if credentials else ""
        base_url = "https://financialmodelingprep.com/stable"
        url = f"{base_url}/key-executives?symbol={query.symbol}&apikey={api_key}"

        return await get_data_many(url, **kwargs)

    @staticmethod
    def transform_data(
        query: FMPKeyExecutivesQueryParams, data: list[dict], **kwargs: Any
    ) -> list[FMPKeyExecutivesData]:
        """Return the transformed data."""
        return [FMPKeyExecutivesData.model_validate(d) for d in data]



"""FMP ETF Sectors Model."""

# pylint: disable=unused-argument

from typing import Any, Optional

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.etf_sectors import (
    EtfSectorsData,
    EtfSectorsQueryParams,
)
from openbb_core.provider.utils.errors import EmptyDataError
from pydantic import field_validator


class FMPEtfSectorsQueryParams(EtfSectorsQueryParams):
    """FMP ETF Sectors Query."""

    __json_schema_extra__ = {"symbol": {"multiple_items_allowed": True}}


class FMPEtfSectorsData(EtfSectorsData):
    """FMP ETF Sectors Data."""

    __alias_dict__ = {"weight": "weightPercentage"}

    @field_validator("weight", mode="before", check_fields=False)
    @classmethod
    def _normalize_percent(cls, v):
        """Normalize percent values."""
        return float(v) / 100 if v else None


class FMPEtfSectorsFetcher(
    Fetcher[
        FMPEtfSectorsQueryParams,
        list[FMPEtfSectorsData],
    ]
):
    """FMP ETF Sectors Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPEtfSectorsQueryParams:
        """Transform the query."""
        return FMPEtfSectorsQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPEtfSectorsQueryParams,
        credentials: Optional[dict[str, str]],
        **kwargs: Any,
    ) -> list:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        from openbb_fmp.utils.helpers import get_data_urls

        api_key = credentials.get("fmp_api_key") if credentials else ""
        symbols = query.symbol.split(",")

        urls = [
            f"https://financialmodelingprep.com/stable/etf/sector-weightings?symbol={symbol.upper()}&apikey={api_key}"
            for symbol in symbols
        ]

        return await get_data_urls(urls, **kwargs)  # type: ignore

    @staticmethod
    def transform_data(
        query: FMPEtfSectorsQueryParams, data: list, **kwargs: Any
    ) -> list[FMPEtfSectorsData]:
        """Return the transformed data."""
        # pylint: disable=import-outside-toplevel
        import warnings

        if not data:
            raise EmptyDataError("No data found")

        symbols = set(query.symbol.split(","))
        returned_symbols = {
            d.get("symbol", "").upper() for d in data if d.get("symbol")
        }
        missing_symbols = symbols - returned_symbols

        if missing_symbols:
            warnings.warn(f"Missing symbols in response: {missing_symbols}")

        return [
            FMPEtfSectorsData.model_validate(d)
            for d in sorted(
                data,
                key=lambda x: (
                    query.symbol.split(",").index(x.get("symbol", len(symbols))),
                    -(x.get("weightPercentage", 0) or 0),
                ),
            )
        ]



"""FMP Available Indices Model."""

# pylint: disable=unused-argument

from typing import Any, Optional

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.available_indices import (
    AvailableIndicesData,
    AvailableIndicesQueryParams,
)


class FMPAvailableIndicesQueryParams(AvailableIndicesQueryParams):
    """FMP Available Indices Query.

    Source: https://site.financialmodelingprep.com/developer/docs#indexes-list
    """


class FMPAvailableIndicesData(AvailableIndicesData):
    """FMP Available Indices Data."""


class FMPAvailableIndicesFetcher(
    Fetcher[
        FMPAvailableIndicesQueryParams,
        list[FMPAvailableIndicesData],
    ]
):
    """FMP Available Indices Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPAvailableIndicesQueryParams:
        """Transform the query params."""
        return FMPAvailableIndicesQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPAvailableIndicesQueryParams,
        credentials: Optional[dict[str, str]],
        **kwargs: Any,
    ) -> list[dict]:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        from openbb_fmp.utils.helpers import get_data_many

        api_key = credentials.get("fmp_api_key") if credentials else ""
        url = f"https://financialmodelingprep.com/stable/index-list?apikey={api_key}"

        return await get_data_many(url, **kwargs)

    @staticmethod
    def transform_data(
        query: FMPAvailableIndicesQueryParams, data: list[dict], **kwargs: Any
    ) -> list[FMPAvailableIndicesData]:
        """Return the transformed data."""
        return [FMPAvailableIndicesData.model_validate(d) for d in data]



"""FMP ETF Equity Exposure Model."""

# pylint: disable=unused-argument

from typing import Any, Optional

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.etf_equity_exposure import (
    EtfEquityExposureData,
    EtfEquityExposureQueryParams,
)
from openbb_core.provider.utils.errors import EmptyDataError
from pydantic import field_validator


class FMPEtfEquityExposureQueryParams(EtfEquityExposureQueryParams):
    """
    FMP ETF Equity Exposure Query Params.

    Source: https://site.financialmodelingprep.com/developer/docs#etf-asset-exposure
    """

    __json_schema_extra__ = {"symbol": {"multiple_items_allowed": True}}


class FMPEtfEquityExposureData(EtfEquityExposureData):
    """FMP ETF Equity Exposure Data."""

    __alias_dict__ = {
        "equity_symbol": "asset",
        "etf_symbol": "symbol",
        "shares": "sharesNumber",
        "weight": "weightPercentage",
        "market_value": "marketValue",
    }

    @field_validator("weight", mode="before", check_fields=False)
    @classmethod
    def normalize_percent(cls, v):
        """Normalize percent values."""
        return float(v) / 100 if v else None


class FMPEtfEquityExposureFetcher(
    Fetcher[FMPEtfEquityExposureQueryParams, list[FMPEtfEquityExposureData]]
):
    """FMP ETF Equity Exposure Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPEtfEquityExposureQueryParams:
        """Transform the query."""
        return FMPEtfEquityExposureQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPEtfEquityExposureQueryParams,
        credentials: Optional[dict[str, str]],
        **kwargs: Any,
    ) -> list[dict]:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        import asyncio  # noqa
        import warnings
        from openbb_fmp.utils.helpers import get_data

        api_key = credentials.get("fmp_api_key") if credentials else ""
        symbols = query.symbol.split(",")
        results: list[dict] = []

        async def get_one(symbol):
            """Get one symbol."""
            url = f"https://financialmodelingprep.com/stable/etf/asset-exposure?symbol={symbol}&apikey={api_key}"
            response = await get_data(url, **kwargs)
            if not response:
                warnings.warn(f"No results found for {symbol}.")
            if response:
                results.extend(response)

        await asyncio.gather(*[get_one(symbol) for symbol in symbols])

        if not results:
            raise EmptyDataError("No data was found for the given symbols.")

        return results

    @staticmethod
    def transform_data(
        query: FMPEtfEquityExposureQueryParams,
        data: list[dict],
        **kwargs: Any,
    ) -> list[FMPEtfEquityExposureData]:
        """Return the transformed data."""
        return [
            FMPEtfEquityExposureData.model_validate(d)
            for d in sorted(data, key=lambda x: x["marketValue"], reverse=True)
        ]



"""FMP Equity Quote Model."""

# pylint: disable=unused-argument

from datetime import date as dateType
from typing import Any, Optional, Union

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.equity_quote import (
    EquityQuoteData,
    EquityQuoteQueryParams,
)
from openbb_core.provider.utils.errors import EmptyDataError
from pydantic import Field, field_validator


class FMPEquityQuoteQueryParams(EquityQuoteQueryParams):
    """FMP Equity Quote Query.

    Source: https://site.financialmodelingprep.com/developer/docs#quote
    """

    __json_schema_extra__ = {"symbol": {"multiple_items_allowed": True}}


class FMPEquityQuoteData(EquityQuoteData):
    """FMP Equity Quote Data."""

    __alias_dict__ = {
        "ma50": "priceAvg50",
        "ma200": "priceAvg200",
        "last_timestamp": "timestamp",
        "high": "dayHigh",
        "low": "dayLow",
        "last_price": "price",
        "change_percent": "changePercentage",
        "prev_close": "previousClose",
    }
    ma50: Optional[float] = Field(
        default=None, description="50 day moving average price."
    )
    ma200: Optional[float] = Field(
        default=None, description="200 day moving average price."
    )
    market_cap: Optional[float] = Field(
        default=None, description="Market cap of the company."
    )

    @field_validator("last_timestamp", mode="before", check_fields=False)
    @classmethod
    def validate_last_timestamp(cls, v: Union[str, int]) -> Optional[dateType]:
        """Return the date as a datetime object."""
        # pylint: disable=import-outside-toplevel
        from datetime import timezone  # noqa
        from openbb_core.provider.utils.helpers import safe_fromtimestamp

        if v:
            v = int(v) if isinstance(v, str) else v
            return safe_fromtimestamp(v, tz=timezone.utc).replace(tzinfo=None)
        return None

    @field_validator("change_percent", mode="after", check_fields=False)
    @classmethod
    def _normalize_percent(cls, v):  # pylint: disable=E0213
        """Return the percent value as a normalized value."""
        return float(v) / 100 if v else None


class FMPEquityQuoteFetcher(
    Fetcher[
        FMPEquityQuoteQueryParams,
        list[FMPEquityQuoteData],
    ]
):
    """FMP Equity Quote Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPEquityQuoteQueryParams:
        """Transform the query params."""
        return FMPEquityQuoteQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPEquityQuoteQueryParams,
        credentials: Optional[dict[str, str]],
        **kwargs: Any,
    ) -> list[dict]:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        import asyncio  # noqa
        import warnings
        from openbb_core.provider.utils.helpers import amake_request
        from openbb_fmp.utils.helpers import response_callback

        api_key = credentials.get("fmp_api_key") if credentials else ""
        base_url = "https://financialmodelingprep.com/stable/quote?"
        symbols = query.symbol.split(",")
        results: list = []

        async def get_one(symbol):
            """Get data for one symbol."""
            url = f"{base_url}symbol={symbol}&apikey={api_key}"
            result = await amake_request(
                url, response_callback=response_callback, **kwargs
            )
            if not result or len(result) == 0:
                warnings.warn(f"Symbol Error: No data found for {symbol}")
            if result and len(result) > 0:
                results.extend(result)

        await asyncio.gather(*[get_one(s) for s in symbols])

        if not results:
            raise EmptyDataError("No data found for the given symbols.")

        return sorted(
            results,
            key=(lambda item: (symbols.index(item.get("symbol", len(symbols))))),
        )

    @staticmethod
    def transform_data(
        query: FMPEquityQuoteQueryParams, data: list[dict], **kwargs: Any
    ) -> list[FMPEquityQuoteData]:
        """Return the transformed data."""
        return [FMPEquityQuoteData.model_validate(d) for d in data]



"""FMP Economic Calendar Model."""

# pylint: disable=unused-argument

from datetime import datetime, timedelta
from typing import Any, Optional

from openbb_core.app.model.abstract.error import OpenBBError
from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.economic_calendar import (
    EconomicCalendarData,
    EconomicCalendarQueryParams,
)
from openbb_core.provider.utils.errors import EmptyDataError, UnauthorizedError
from pydantic import Field, field_validator, model_validator


class FMPEconomicCalendarQueryParams(EconomicCalendarQueryParams):
    """FMP Economic Calendar Query.

    Source: https://site.financialmodelingprep.com/developer/docs#economics-calendar
    """


class FMPEconomicCalendarData(EconomicCalendarData):
    """FMP Economics Calendar Data.

    Source: https://site.financialmodelingprep.com/developer/docs#economics-calendar
    """

    __alias_dict__ = {
        "consensus": "estimate",
        "importance": "impact",
        "last_updated": "updatedAt",
        "created_at": "createdAt",
        "change_percent": "changePercentage",
    }

    change: Optional[float] = Field(
        description="Value change since previous.",
        default=None,
    )
    change_percent: Optional[float] = Field(
        description="Percentage change since previous.",
        default=None,
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )

    @field_validator("change_percent", mode="before", check_fields=False)
    @classmethod
    def _normalize_percent(cls, v):
        """Normalize percentage value."""
        return v / 100 if v else None

    @model_validator(mode="before")
    @classmethod
    def empty_strings(cls, values):
        """Replace empty values with None."""
        return (
            {k: (None if v in ("", 0) else v) for k, v in values.items()}
            if isinstance(values, dict)
            else values
        )


class FMPEconomicCalendarFetcher(
    Fetcher[
        FMPEconomicCalendarQueryParams,
        list[FMPEconomicCalendarData],
    ]
):
    """FMP Economic Calendar Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPEconomicCalendarQueryParams:
        """Transform the query."""
        transformed_params = params
        if not transformed_params.get("start_date"):
            transformed_params["start_date"] = (
                datetime.now() - timedelta(days=1)
            ).date()
        if not transformed_params.get("end_date"):
            transformed_params["end_date"] = (datetime.now() + timedelta(days=7)).date()
        return FMPEconomicCalendarQueryParams(**transformed_params)

    @staticmethod
    async def aextract_data(
        query: FMPEconomicCalendarQueryParams,
        credentials: Optional[dict[str, str]],
        **kwargs: Any,
    ) -> list[dict]:
        """Return the data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        import asyncio  # noqa
        import warnings
        from openbb_core.provider.utils.helpers import amake_request
        from openbb_fmp.utils.helpers import response_callback

        api_key = credentials.get("fmp_api_key") if credentials else ""

        base_url = "https://financialmodelingprep.com/stable/economic-calendar?"

        # FMP allows only 3-month windows to be queried, we need to chunk to request.
        def date_range(start_date, end_date):
            """Yield start and end dates for each 90-day period between start_date and end_date."""
            delta = timedelta(days=90)
            current_date = start_date
            while current_date < end_date:
                next_date = min(current_date + delta, end_date)
                yield current_date, next_date
                current_date = next_date + timedelta(days=1)

        date_ranges = list(date_range(query.start_date, query.end_date))
        urls = [
            f"{base_url}from={start_date.strftime('%Y-%m-%d')}"  # type: ignore
            + f"&to={end_date.strftime('%Y-%m-%d')}&apikey={api_key}"
            for start_date, end_date in date_ranges
        ]
        results: list[dict] = []

        # We need to do this because Pytest does not seem to be able to handle `amake_requests`.
        async def get_one(url):
            """Get data for one URL."""
            n_urls = 1
            try:
                result = await amake_request(url, response_callback=response_callback)
                if result:
                    results.extend(result)
            except UnauthorizedError as e:
                raise e from e
            except OpenBBError as e:
                if len(urls) == 1 or (len(urls) > 1 and n_urls == len(urls)):
                    raise e from e
                warnings.warn(f"Error in fetching part of the data from FMP -> {e}")
            n_urls += 1

        await asyncio.gather(*[get_one(url) for url in urls])

        if not results:
            raise EmptyDataError("The request was returned empty.")

        return results

    @staticmethod
    def transform_data(
        query: FMPEconomicCalendarQueryParams,
        data: list[dict],
        **kwargs: Any,
    ) -> list[FMPEconomicCalendarData]:
        """Transform the data."""
        return [FMPEconomicCalendarData.model_validate(d) for d in data]



"""FMP ETF Countries Model."""

# pylint: disable=unused-argument

from typing import Any, Optional

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.etf_countries import (
    EtfCountriesData,
    EtfCountriesQueryParams,
)


class FMPEtfCountriesQueryParams(EtfCountriesQueryParams):
    """FMP ETF Countries Query."""

    __json_schema_extra__ = {"symbol": {"multiple_items_allowed": True}}


class FMPEtfCountriesData(EtfCountriesData):
    """FMP ETF Countries Data."""


class FMPEtfCountriesFetcher(
    Fetcher[
        FMPEtfCountriesQueryParams,
        list[FMPEtfCountriesData],
    ]
):
    """FMP ETF Countries Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPEtfCountriesQueryParams:
        """Transform the query."""
        return FMPEtfCountriesQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPEtfCountriesQueryParams,
        credentials: Optional[dict[str, str]],
        **kwargs: Any,
    ) -> list:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        import asyncio  # noqa
        import warnings
        from openbb_core.provider.utils.errors import EmptyDataError
        from openbb_fmp.utils.helpers import get_data

        api_key = credentials.get("fmp_api_key") if credentials else ""
        symbols = query.symbol.split(",")
        results: list = []

        async def get_one(symbol):
            """Get data for one symbol."""
            url = f"https://financialmodelingprep.com/stable/etf/country-weightings?symbol={symbol}&apikey={api_key}"
            result = await get_data(url, **kwargs)

            if not result:
                warnings.warn(f"Symbol Error: No data found for {symbol}")
                return

            new_data: list = []
            if result:
                for row in result:
                    row_weight = row.get("weightPercentage", "0%").replace("%", "")
                    if not row_weight or row_weight == "0":
                        continue
                    new_row = {
                        "symbol": symbol,
                        "country": row["country"],
                        "weight": float(row_weight) * 0.01,
                    }
                    new_data.append(new_row)

                if new_data:
                    results.extend(new_data)
            return

        await asyncio.gather(*[get_one(symbol) for symbol in symbols])

        if not results:
            raise EmptyDataError("No data found for the given symbols.")

        return results

    @staticmethod
    def transform_data(
        query: FMPEtfCountriesQueryParams, data: list, **kwargs: Any
    ) -> list[FMPEtfCountriesData]:
        """Return the transformed data."""
        symbols = query.symbol.split(",")

        return [
            FMPEtfCountriesData.model_validate(d)
            for d in sorted(
                data,
                key=lambda x: (
                    symbols.index(x.get("symbol", len(symbols))),
                    -(x.get("weightPercentage", 0) or 0),
                ),
            )
        ]



"""FMP ETF Holdings Model."""

# pylint: disable=unused-argument

from datetime import (
    date as dateType,
    datetime,
)
from typing import Any, Optional, Union

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.etf_holdings import (
    EtfHoldingsData,
    EtfHoldingsQueryParams,
)
from openbb_core.provider.utils.errors import EmptyDataError
from pydantic import ConfigDict, Field, field_validator


class FMPEtfHoldingsQueryParams(EtfHoldingsQueryParams):
    """FMP ETF Holdings Query.

    Source: https://site.financialmodelingprep.com/developer/docs#holdings
    """


class FMPEtfHoldingsData(EtfHoldingsData):
    """FMP ETF Holdings Data."""

    model_config = ConfigDict(extra="ignore")

    __alias_dict__ = {
        "shares": "sharesNumber",
        "value": "marketValue",
        "weight": "weightPercentage",
        "symbol": "asset",
        "updated": "updatedAt",
        "cusip": "securityCusip",
    }

    cusip: Optional[str] = Field(
        description="The CUSIP of the holding.", default=None, title="CUSIP"
    )
    isin: Optional[str] = Field(
        description="The ISIN of the holding.", default=None, title="ISIN"
    )
    weight: Optional[float] = Field(
        description="The weight of the holding, as a normalized percent.",
        default=None,
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    shares: Optional[Union[float, str]] = Field(
        description="The number of shares held.", default=None
    )
    value: Optional[float] = Field(
        description="The market value of the holding.",
        default=None,
        json_schema_extra={"x-unit_measurement": "currency"},
    )
    updated: Optional[Union[dateType, datetime]] = Field(
        description="The date the data was updated.", default=None
    )

    @field_validator("weight", mode="before", check_fields=False)
    @classmethod
    def normalize_percent(cls, v):
        """Normalize percent values."""
        return float(v) / 100 if v else None

    @field_validator(
        "cusip", "isin", "name", "asset", "value", mode="before", check_fields=False
    )
    @classmethod
    def replace_empty(cls, v):
        """Replace empty strings and 0s with None."""
        if isinstance(v, str):
            return v if v not in ("", "0", "-") else None
        return v if v and v != 0 else None


class FMPEtfHoldingsFetcher(
    Fetcher[
        FMPEtfHoldingsQueryParams,
        list[FMPEtfHoldingsData],
    ]
):
    """FMP ETF Holdings Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPEtfHoldingsQueryParams:
        """Transform the query."""
        return FMPEtfHoldingsQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPEtfHoldingsQueryParams,
        credentials: Optional[dict[str, str]],
        **kwargs: Any,
    ) -> list[dict]:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        from openbb_fmp.utils.helpers import get_data_many

        api_key = credentials.get("fmp_api_key") if credentials else ""
        url = f"https://financialmodelingprep.com/stable/etf/holdings?symbol={query.symbol}&apikey={api_key}"
        return await get_data_many(url, **kwargs)

    @staticmethod
    def transform_data(
        query: FMPEtfHoldingsQueryParams,
        data: list[dict],
        **kwargs: Any,
    ) -> list[FMPEtfHoldingsData]:
        """Return the transformed data."""
        if not data:
            raise EmptyDataError("No data was found for the given symbol.")
        return [FMPEtfHoldingsData.model_validate(item) for item in data]



"""FMP Cash Flow Statement Model."""

# pylint: disable=unused-argument

from datetime import (
    date as dateType,
    datetime,
)
from typing import Any, Optional

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.cash_flow import (
    CashFlowStatementData,
    CashFlowStatementQueryParams,
)
from openbb_core.provider.utils.descriptions import QUERY_DESCRIPTIONS
from openbb_fmp.utils.definitions import FinancialStatementPeriods
from pydantic import Field


class FMPCashFlowStatementQueryParams(CashFlowStatementQueryParams):
    """FMP Cash Flow Statement Query.

    Source: https://financialmodelingprep.com/developer/docs/#Cash-Flow-Statement
    """

    period: FinancialStatementPeriods = Field(
        default="annual",
        description=QUERY_DESCRIPTIONS.get("period", ""),
    )


class FMPCashFlowStatementData(CashFlowStatementData):
    """FMP Cash Flow Statement Data."""

    __alias_dict__ = {
        "period_ending": "date",
        "fiscal_period": "period",
        "fiscal_year": "calendarYear",
        "filing_date": "fillingDate",
        "accepted_date": "acceptedDate",
        "reported_currency": "reportedCurrency",
        "net_income": "netIncome",
        "depreciation_and_amortization": "depreciationAndAmortization",
        "deferred_income_tax": "deferredIncomeTax",
        "stock_based_compensation": "stockBasedCompensation",
        "change_in_working_capital": "changeInWorkingCapital",
        "change_in_account_receivables": "accountsReceivables",
        "change_in_inventory": "inventory",
        "change_in_account_payable": "accountsPayables",
        "change_in_other_working_capital": "otherWorkingCapital",
        "change_in_other_non_cash_items": "otherNonCashItems",
        "net_cash_from_operating_activities": "netCashProvidedByOperatingActivities",
        "purchase_of_property_plant_and_equipment": "investmentsInPropertyPlantAndEquipment",
        "acquisitions": "acquisitionsNet",
        "purchase_of_investment_securities": "purchasesOfInvestments",
        "sale_and_maturity_of_investments": "salesMaturitiesOfInvestments",
        "other_investing_activities": "otherInvestingActivities",
        "net_cash_from_investing_activities": "netCashProvidedByInvestingActivities",
        "repayment_of_debt": "debtRepayment",
        "issuance_of_common_equity": "commonStockIssuance",
        "repurchase_of_common_equity": "commonStockRepurchased",
        "net_common_equity_issuance": "netCommonStockIssuance",
        "net_preferred_equity_issuance": "netPreferredStockIssuance",
        "net_equity_issuance": "netStockIssuance",
        "payment_of_dividends": "dividendsPaid",
        "other_financing_activities": "otherFinancingActivites",
        "net_cash_from_financing_activities": "netCashProvidedByFinancingActivities",
        "effect_of_exchange_rate_changes_on_cash": "effectOfForexChangesOnCash",
        "net_change_in_cash_and_equivalents": "netChangeInCash",
        "cash_at_beginning_of_period": "cashAtBeginningOfPeriod",
        "cash_at_end_of_period": "cashAtEndOfPeriod",
        "operating_cash_flow": "operatingCashFlow",
        "capital_expenditure": "capitalExpenditure",
        "free_cash_flow": "freeCashFlow",
    }

    fiscal_year: Optional[int] = Field(
        default=None,
        description="The fiscal year of the fiscal period.",
    )
    filing_date: Optional[dateType] = Field(
        default=None,
        description="The date of the filing.",
    )
    accepted_date: Optional[datetime] = Field(
        default=None, description="The date the filing was accepted."
    )
    cik: Optional[str] = Field(
        default=None,
        description="The Central Index Key (CIK) assigned by the SEC, if applicable.",
    )
    symbol: Optional[str] = Field(
        default=None,
        description="The stock ticker symbol.",
    )
    reported_currency: Optional[str] = Field(
        default=None,
        description="The currency in which the cash flow statement was reported.",
    )
    net_income: Optional[int] = Field(
        default=None,
        description="Net income.",
    )
    depreciation_and_amortization: Optional[int] = Field(
        default=None,
        description="Depreciation and amortization.",
    )
    deferred_income_tax: Optional[int] = Field(
        default=None,
        description="Deferred income tax.",
    )
    stock_based_compensation: Optional[int] = Field(
        default=None,
        description="Stock-based compensation.",
    )
    change_in_working_capital: Optional[int] = Field(
        default=None,
        description="Change in working capital.",
    )
    change_in_account_receivables: Optional[int] = Field(
        default=None,
        description="Change in account receivables.",
    )
    change_in_inventory: Optional[int] = Field(
        default=None,
        description="Change in inventory.",
    )
    change_in_account_payable: Optional[int] = Field(
        default=None,
        description="Change in account payable.",
    )
    change_in_other_working_capital: Optional[int] = Field(
        default=None,
        description="Change in other working capital.",
    )
    change_in_other_non_cash_items: Optional[int] = Field(
        default=None,
        description="Change in other non-cash items.",
    )
    net_cash_from_operating_activities: Optional[int] = Field(
        default=None,
        description="Net cash from operating activities.",
    )
    purchase_of_property_plant_and_equipment: Optional[int] = Field(
        default=None,
        description="Purchase of property, plant and equipment.",
    )
    acquisitions: Optional[int] = Field(
        default=None,
        description="Acquisitions.",
    )
    purchase_of_investment_securities: Optional[int] = Field(
        default=None,
        description="Purchase of investment securities.",
    )
    sale_and_maturity_of_investments: Optional[int] = Field(
        default=None,
        description="Sale and maturity of investments.",
    )
    other_investing_activities: Optional[int] = Field(
        default=None,
        description="Other investing activities.",
    )
    net_cash_from_investing_activities: Optional[int] = Field(
        default=None,
        description="Net cash from investing activities.",
    )
    repayment_of_debt: Optional[int] = Field(
        default=None,
        description="Repayment of debt.",
    )
    issuance_of_common_equity: Optional[int] = Field(
        default=None,
        description="Issuance of common equity.",
    )
    repurchase_of_common_equity: Optional[int] = Field(
        default=None,
        description="Repurchase of common equity.",
    )
    net_common_equity_issuance: Optional[int] = Field(
        default=None,
        description="Net common equity issuance.",
    )
    net_preferred_equity_issuance: Optional[int] = Field(
        default=None,
        description="Net preferred equity issuance.",
    )
    net_equity_issuance: Optional[int] = Field(
        default=None,
        description="Net equity issuance.",
    )
    short_term_net_debt_issuance: Optional[int] = Field(
        default=None,
        description="Short term net debt issuance.",
    )
    long_term_net_debt_issuance: Optional[int] = Field(
        default=None,
        description="Long term net debt issuance.",
    )
    net_debt_issuance: Optional[int] = Field(
        default=None,
        description="Net debt issuance.",
    )
    common_dividends_paid: Optional[int] = Field(
        default=None,
        description="Payment of common dividends.",
    )
    preferred_dividends_paid: Optional[int] = Field(
        default=None,
        description="Payment of preferred dividends.",
    )
    net_dividends_paid: Optional[int] = Field(
        default=None,
        description="Net dividends paid.",
    )
    other_financing_activities: Optional[int] = Field(
        default=None,
        description="Other financing activities.",
    )
    net_cash_from_financing_activities: Optional[int] = Field(
        default=None,
        description="Net cash from financing activities.",
    )
    effect_of_exchange_rate_changes_on_cash: Optional[int] = Field(
        default=None,
        description="Effect of exchange rate changes on cash.",
    )
    net_change_in_cash_and_equivalents: Optional[int] = Field(
        default=None,
        description="Net change in cash and equivalents.",
    )
    cash_at_beginning_of_period: Optional[int] = Field(
        default=None,
        description="Cash at beginning of period.",
    )
    cash_at_end_of_period: Optional[int] = Field(
        default=None,
        description="Cash at end of period.",
    )
    operating_cash_flow: Optional[int] = Field(
        default=None,
        description="Operating cash flow.",
    )
    capital_expenditure: Optional[int] = Field(
        default=None,
        description="Capital expenditure.",
    )
    income_taxes_paid: Optional[int] = Field(
        default=None,
        description="Income taxes paid.",
    )
    interest_paid: Optional[int] = Field(
        default=None,
        description="Interest paid.",
    )
    free_cash_flow: Optional[int] = Field(
        default=None,
    )


class FMPCashFlowStatementFetcher(
    Fetcher[
        FMPCashFlowStatementQueryParams,
        list[FMPCashFlowStatementData],
    ]
):
    """Transform the query, extract and transform the data from the FMP endpoints."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPCashFlowStatementQueryParams:
        """Transform the query params."""
        return FMPCashFlowStatementQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPCashFlowStatementQueryParams,
        credentials: Optional[dict[str, str]],
        **kwargs: Any,
    ) -> list[dict]:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        from openbb_fmp.utils.helpers import get_data_many

        api_key = credentials.get("fmp_api_key") if credentials else ""
        base_url = "https://financialmodelingprep.com/stable/cash-flow-statement"

        if query.period == "ttm":
            base_url += "-ttm"

        url = (
            base_url
            + f"?symbol={query.symbol}{'&period=' + query.period if query.period != 'ttm' else ''}"
            + f"&limit={query.limit if query.limit else 5}"
            + f"&apikey={api_key}"
        )

        return await get_data_many(url, **kwargs)

    @staticmethod
    def transform_data(
        query: FMPCashFlowStatementQueryParams, data: list[dict], **kwargs: Any
    ) -> list[FMPCashFlowStatementData]:
        """Return the transformed data."""
        return [FMPCashFlowStatementData.model_validate(d) for d in data]



"""FMP Risk Premium Model."""

# pylint: disable=unused-argument

from typing import Any, Optional

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.risk_premium import (
    RiskPremiumData,
    RiskPremiumQueryParams,
)


class FMPRiskPremiumQueryParams(RiskPremiumQueryParams):
    """FMP Risk Premium Query.

    Source: https://site.financialmodelingprep.com/developer/docs#market-risk-premium
    """


class FMPRiskPremiumData(RiskPremiumData):
    """FMP Risk Premium Data."""


class FMPRiskPremiumFetcher(
    Fetcher[
        FMPRiskPremiumQueryParams,
        list[FMPRiskPremiumData],
    ]
):
    """FMP Risk Premium Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPRiskPremiumQueryParams:
        """Transform the query params."""
        return FMPRiskPremiumQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPRiskPremiumQueryParams,
        credentials: Optional[dict[str, str]],
        **kwargs: Any,
    ) -> list[dict]:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        from openbb_fmp.utils.helpers import get_data_many

        api_key = credentials.get("fmp_api_key") if credentials else ""
        url = f"https://financialmodelingprep.com/stable/market-risk-premium?apikey={api_key}"

        return await get_data_many(url, **kwargs)

    @staticmethod
    def transform_data(
        query: FMPRiskPremiumQueryParams, data: list[dict], **kwargs: Any
    ) -> list[FMPRiskPremiumData]:
        """Return the transformed data."""
        return [FMPRiskPremiumData.model_validate(d) for d in data]



"""FMP Currency Available Pairs Model."""

# pylint: disable=unused-argument

from typing import Any, Dict, List, Optional

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.currency_pairs import (
    CurrencyPairsData,
    CurrencyPairsQueryParams,
)
from openbb_core.provider.utils.errors import EmptyDataError
from pydantic import Field


class FMPCurrencyPairsQueryParams(CurrencyPairsQueryParams):
    """FMP Currency Available Pairs Query.

    Source: https://site.financialmodelingprep.com/developer/docs#forex
    """


class FMPCurrencyPairsData(CurrencyPairsData):
    """FMP Currency Available Pairs Data."""

    symbol: str = Field(description="Symbol of the currency pair.")
    from_currency: str = Field(description="Base currency of the currency pair.")
    to_currency: str = Field(description="Quote currency of the currency pair.")
    from_name: str = Field(description="Name of the base currency.")
    to_name: str = Field(description="Name of the quote currency.")


class FMPCurrencyPairsFetcher(
    Fetcher[
        FMPCurrencyPairsQueryParams,
        List[FMPCurrencyPairsData],
    ]
):
    """Transform the query, extract and transform the data from the FMP endpoints."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> FMPCurrencyPairsQueryParams:
        """Transform the query params."""
        return FMPCurrencyPairsQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPCurrencyPairsQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        from openbb_fmp.utils.helpers import get_data_many

        api_key = credentials.get("fmp_api_key") if credentials else ""
        url = f"https://financialmodelingprep.com/stable/forex-list?apikey={api_key}"

        return await get_data_many(url, **kwargs)

    @staticmethod
    def transform_data(
        query: FMPCurrencyPairsQueryParams, data: List[Dict], **kwargs: Any
    ) -> List[FMPCurrencyPairsData]:
        """Return the transformed data."""
        # pylint: disable=import-outside-toplevel
        from pandas import DataFrame

        if not data:
            raise EmptyDataError("The request was returned empty.")

        df = DataFrame(data)

        if query.query:
            df = df[
                df["symbol"].str.contains(query.query, case=False)
                | df["fromCurrency"].str.contains(query.query, case=False)
                | df["toCurrency"].str.contains(query.query, case=False)
                | df["fromName"].str.contains(query.query, case=False)
                | df["toName"].str.contains(query.query, case=False)
            ]

        if len(df) == 0:
            raise EmptyDataError(
                f"No results were found with the query supplied. -> {query.query}"
            )
        return [FMPCurrencyPairsData.model_validate(d) for d in df.to_dict("records")]



"""FMP Share Statistics Model."""

# pylint: disable=unused-argument

from typing import Any, Optional

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.share_statistics import (
    ShareStatisticsData,
    ShareStatisticsQueryParams,
)
from pydantic import Field, field_validator


class FMPShareStatisticsQueryParams(ShareStatisticsQueryParams):
    """FMP Share Statistics Query.

    Source: https://site.financialmodelingprep.com/developer/docs#shares-float
    """

    __json_schema_extra__ = {"symbol": {"multiple_items_allowed": True}}


class FMPShareStatisticsData(ShareStatisticsData):
    """FMP Share Statistics Data."""

    __alias_dict__ = {"url": "source"}

    url: Optional[str] = Field(
        default=None,
        description="URL to the source document, if available.",
    )

    @field_validator("free_float", mode="before")
    @classmethod
    def _normalize_percent(cls, v):
        """Return the date as a datetime object."""
        return v / 100 if v else None


class FMPShareStatisticsFetcher(
    Fetcher[
        FMPShareStatisticsQueryParams,
        list[FMPShareStatisticsData],
    ]
):
    """FMP Share Statistics Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPShareStatisticsQueryParams:
        """Transform the query params."""
        return FMPShareStatisticsQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPShareStatisticsQueryParams,
        credentials: Optional[dict[str, str]],
        **kwargs: Any,
    ) -> list:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        from openbb_fmp.utils.helpers import get_data_urls

        api_key = credentials.get("fmp_api_key") if credentials else ""
        symbols = query.symbol.split(",")

        urls = [
            f"https://financialmodelingprep.com/stable/shares-float?symbol={symbol}&apikey={api_key}"
            for symbol in symbols
        ]

        return await get_data_urls(urls, **kwargs)  # type: ignore

    @staticmethod
    def transform_data(
        query: FMPShareStatisticsQueryParams, data: list, **kwargs: Any
    ) -> list[FMPShareStatisticsData]:
        """Return the transformed data."""
        # pylint: disable=import-outside-toplevel
        import warnings

        symbols = query.symbol.split(",")

        if len(symbols) != len(data):
            data_symbols = {d["symbol"] for d in data if "symbol" in d}
            missing_symbols = set(symbols) - data_symbols
            if missing_symbols:
                warnings.warn(
                    f"No data found for symbols: {', '.join(missing_symbols)}"
                )

        return [
            FMPShareStatisticsData.model_validate(d)
            for d in sorted(
                data,
                key=(lambda item: (symbols.index(item.get("symbol", len(symbols))))),
            )
        ]



"""FMP Revenue By Business Line Model."""

# pylint: disable=unused-argument

from typing import Any, Literal, Optional

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.revenue_business_line import (
    RevenueBusinessLineData,
    RevenueBusinessLineQueryParams,
)
from openbb_core.provider.utils.descriptions import QUERY_DESCRIPTIONS
from openbb_core.provider.utils.errors import EmptyDataError
from pydantic import Field


class FMPRevenueBusinessLineQueryParams(RevenueBusinessLineQueryParams):
    """FMP Revenue By Business Line Query.

    Source: https://site.financialmodelingprep.com/developer/docs#revenue-product-segmentation
    """

    period: Literal["quarter", "annual"] = Field(
        default="annual", description=QUERY_DESCRIPTIONS.get("period", "")
    )


class FMPRevenueBusinessLineData(RevenueBusinessLineData):
    """FMP Revenue By Business Line Data."""


class FMPRevenueBusinessLineFetcher(
    Fetcher[
        FMPRevenueBusinessLineQueryParams,
        list[FMPRevenueBusinessLineData],
    ]
):
    """FMP Revenue By Business Line Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPRevenueBusinessLineQueryParams:
        """Transform the query params."""
        return FMPRevenueBusinessLineQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPRevenueBusinessLineQueryParams,
        credentials: Optional[dict[str, str]],
        **kwargs: Any,
    ) -> list[dict]:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        from openbb_fmp.utils.helpers import get_data_many

        api_key = credentials.get("fmp_api_key") if credentials else ""
        base_url = (
            "https://financialmodelingprep.com/stable/revenue-product-segmentation?"
        )
        url = (
            f"{base_url}symbol={query.symbol}"
            f"&period={query.period}"
            f"&structure=flat&apikey={api_key}"
        )
        return await get_data_many(url, **kwargs)

    @staticmethod
    def transform_data(
        query: FMPRevenueBusinessLineQueryParams,
        data: list[dict],
        **kwargs: Any,
    ) -> list[FMPRevenueBusinessLineData]:
        """Return the transformed data."""
        if not data:
            raise EmptyDataError("The request was returned empty.")

        results: list = []
        # We need to flatten the data.
        for item in data:
            period_ending = item.get("date")
            fiscal_year = item.get("fiscalYear")
            fiscal_period = item.get("period")
            segment = item.get("data", {})

            for business_line, revenue_value in segment.items():
                if revenue_value is not None:
                    revenue = int(revenue_value) if revenue_value is not None else None
                    if revenue is not None:
                        results.append(
                            FMPRevenueBusinessLineData.model_validate(
                                {
                                    "period_ending": period_ending,
                                    "fiscal_year": fiscal_year,
                                    "fiscal_period": fiscal_period,
                                    "business_line": business_line.strip(),
                                    "revenue": revenue,
                                }
                            )
                        )

        if not results:
            raise EmptyDataError("Unknown error while transforming the data.")

        return sorted(results, key=lambda x: (x.period_ending or "", x.revenue or 0))



"""FMP Equity Screener Model."""

# pylint: disable=unused-argument

from typing import Any, Optional

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.equity_screener import (
    EquityScreenerData,
    EquityScreenerQueryParams,
)
from openbb_core.provider.utils.errors import EmptyDataError
from openbb_fmp.utils.definitions import (
    Countries,
    Exchanges,
    IndustryChoices,
    Sectors,
)
from pydantic import Field, field_validator


class FMPEquityScreenerQueryParams(EquityScreenerQueryParams):
    """FMP Equity Screener Query."""

    __json_schema_extra__ = {
        "industry": {
            "x-widget_config": {
                "options": IndustryChoices,
            },
        },
        "sector": {
            "x-widget_config": {
                "options": [
                    {"label": sector.replace("_", " ").title(), "value": sector}
                    for sector in Sectors.__args__
                ],
            },
        },
        "exchange": {
            "x-widget_config": {
                "options": [
                    {"label": exchange.upper(), "value": exchange}
                    for exchange in Exchanges.__args__
                ],
            },
        },
        "country": {
            "x-widget_config": {
                "options": [
                    {"label": country.upper(), "value": country}
                    for country in Countries.__args__
                ],
            },
        },
    }

    __alias_dict__ = {
        "mktcap_min": "marketCapMoreThan",
        "mktcap_max": "marketCapLowerThan",
        "price_min": "priceMoreThan",
        "price_max": "priceLowerThan",
        "beta_min": "betaMoreThan",
        "beta_max": "betaLowerThan",
        "volume_min": "volumeMoreThan",
        "volume_max": "volumeLowerThan",
        "dividend_min": "dividendMoreThan",
        "dividend_max": "dividendLowerThan",
        "is_active": "isActivelyTrading",
        "is_etf": "isEtf",
        "is_fund": "isFund",
        "all_share_classes": "includeAllShareClasses",
    }

    mktcap_min: Optional[int] = Field(
        default=None,
        description="Filter by market cap greater than this value.",
        title="Mkt Cap Min",
    )
    mktcap_max: Optional[int] = Field(
        default=None,
        description="Filter by market cap less than this value.",
        title="Mkt Cap Max",
    )
    price_min: Optional[float] = Field(
        default=None,
        description="Filter by price greater than this value.",
    )
    price_max: Optional[float] = Field(
        default=None,
        description="Filter by price less than this value.",
    )
    beta_min: Optional[float] = Field(
        default=None,
        description="Filter by a beta greater than this value.",
    )
    beta_max: Optional[float] = Field(
        default=None,
        description="Filter by a beta less than this value.",
    )
    volume_min: Optional[int] = Field(
        default=None,
        description="Filter by volume greater than this value.",
    )
    volume_max: Optional[int] = Field(
        default=None,
        description="Filter by volume less than this value.",
    )
    dividend_min: Optional[float] = Field(
        default=None,
        description="Filter by dividend amount greater than this value.",
    )
    dividend_max: Optional[float] = Field(
        default=None,
        description="Filter by dividend amount less than this value.",
    )
    sector: Optional[Sectors] = Field(
        default=None,
        description="Filter by sector.",
    )
    industry: Optional[str] = Field(
        default=None,
        description="Filter by industry.",
    )
    country: Optional[Countries] = Field(
        default=None, description="Filter by country, as a two-letter country code."
    )
    exchange: Optional[Exchanges] = Field(
        default=None, description="Filter by exchange."
    )
    is_etf: Optional[bool] = Field(
        default=None,
        description="If true, includes ETFs.",
    )
    is_active: Optional[bool] = Field(
        default=None,
        description="If false, returns only inactive tickers.",
    )
    is_fund: Optional[bool] = Field(
        default=None,
        description="If true, includes funds.",
    )
    all_share_classes: Optional[bool] = Field(
        default=None,
        description="If true, includes all share classes of a equity.",
    )
    limit: Optional[int] = Field(
        default=50000, description="Limit the number of results to return."
    )

    @field_validator("industry")
    @classmethod
    def _validate_industry(cls, v):
        """Validate industry."""
        industries = [v["value"] for v in IndustryChoices]
        if v and v not in industries + ["all"]:
            raise ValueError(f"Industry must be one of {', '.join(industries)}")
        return v


class FMPEquityScreenerData(EquityScreenerData):
    """FMP Equity Screener Data."""

    __alias_dict__ = {
        "name": "companyName",
        "market_cap": "marketCap",
        "last_annual_dividend": "lastAnnualDividend",
        "exchange": "exchangeShortName",
        "exchange_name": "exchange",
        "is_etf": "isEtf",
        "actively_trading": "isActivelyTrading",
    }

    market_cap: Optional[int] = Field(
        description="The market cap of ticker.", default=None
    )
    sector: Optional[str] = Field(
        description="The sector the ticker belongs to.",
        default=None,
    )
    industry: Optional[str] = Field(
        description="The industry ticker belongs to.", default=None
    )
    beta: Optional[float] = Field(description="The beta of the ETF.", default=None)
    price: Optional[float] = Field(description="The current price.", default=None)
    last_annual_dividend: Optional[float] = Field(
        description="The last annual amount dividend paid.",
        default=None,
    )
    volume: Optional[int] = Field(
        description="The current trading volume.", default=None
    )
    exchange: Optional[str] = Field(
        description="The exchange code the asset trades on.",
        default=None,
    )
    exchange_name: Optional[str] = Field(
        description="The full name of the primary exchange.",
        default=None,
    )
    country: Optional[str] = Field(
        description="The two-letter country abbreviation where the head office is located.",
        default=None,
    )
    is_etf: Optional[bool] = Field(
        description="Whether the ticker is an ETF.", default=None
    )
    is_fund: Optional[bool] = Field(
        description="Whether the ticker is a fund.", default=None
    )
    actively_trading: Optional[bool] = Field(
        description="Whether the ETF is actively trading.",
        default=None,
    )


class FMPEquityScreenerFetcher(
    Fetcher[
        FMPEquityScreenerQueryParams,
        list[FMPEquityScreenerData],
    ]
):
    """FMP Equity Screener Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPEquityScreenerQueryParams:
        """Transform the query."""
        return FMPEquityScreenerQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPEquityScreenerQueryParams,
        credentials: Optional[dict[str, str]],
        **kwargs: Any,
    ) -> list[dict]:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        from openbb_core.provider.utils.helpers import get_querystring
        from openbb_fmp.utils.helpers import get_data

        api_key = credentials.get("fmp_api_key") if credentials else ""
        sector: str = (
            query.sector.replace("_", " ").title().replace(" ", "%20")
            if query.sector
            else ""
        )
        industry_map = {v["value"]: v["label"] for v in IndustryChoices}
        industry: str = (
            industry_map.get(query.industry, query.industry) if query.industry else ""
        )
        industry = (
            industry.replace(" & ", "%20%26%20")
            .replace(" ", "%20")
            .replace("/", "%2F")
            .replace("-", "%2D")
            .replace(",", "%2C")
        )
        exchange: str = query.exchange.upper() if query.exchange else ""
        country: str = query.country.upper() if query.country else ""
        query.is_active = True if query.is_active is None else query.is_active
        query.is_etf = False if query.is_etf is None else query.is_etf
        query.is_fund = False if query.is_fund is None else query.is_fund
        query.all_share_classes = (
            False if query.all_share_classes is None else query.all_share_classes
        )

        query_dict = query.model_dump(exclude_none=True, by_alias=True)

        if sector:
            query_dict["sector"] = sector
        if industry:
            query_dict["industry"] = industry
        if exchange:
            query_dict["exchange"] = exchange
        if country:
            query_dict["country"] = country

        query_str = (
            get_querystring(query_dict, exclude=["query"])
            .replace("True", "true")
            .replace("False", "false")
        )
        base_url = "https://financialmodelingprep.com/stable/company-screener"
        url = f"{base_url}?{query_str}&apikey={api_key}"

        return await get_data(url, **kwargs)  # type: ignore

    @staticmethod
    def transform_data(
        query: FMPEquityScreenerQueryParams, data: list[dict], **kwargs: Any
    ) -> list[FMPEquityScreenerData]:
        """Return the transformed data."""
        if not data:
            raise EmptyDataError("The request was returned empty.")
        return [
            FMPEquityScreenerData.model_validate(d)
            for d in sorted(data, key=lambda x: x["marketCap"], reverse=True)
        ]



"""FMP Analyst Estimates Model."""

# pylint: disable=unused-argument

from typing import Any, Literal, Optional

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.analyst_estimates import (
    AnalystEstimatesData,
    AnalystEstimatesQueryParams,
)
from openbb_core.provider.utils.descriptions import QUERY_DESCRIPTIONS
from openbb_core.provider.utils.errors import EmptyDataError
from pydantic import Field


class FMPAnalystEstimatesQueryParams(AnalystEstimatesQueryParams):
    """FMP Analyst Estimates Query.

    Source: https://site.financialmodelingprep.com/developer/docs/analyst-estimates-api/
    """

    __json_schema_extra__ = {"symbol": {"multiple_items_allowed": True}}

    period: Literal["quarter", "annual"] = Field(
        default="annual", description=QUERY_DESCRIPTIONS.get("period", "")
    )
    limit: Optional[int] = Field(
        default=None, description=QUERY_DESCRIPTIONS.get("limit", "")
    )
    page: Optional[int] = Field(
        default=None, description="Page number for paginated results. Used with limit."
    )


class FMPAnalystEstimatesData(AnalystEstimatesData):
    """FMP Analyst Estimates Data."""

    __alias_dict__ = {
        "estimated_revenue_low": "revenueLow",
        "estimated_revenue_high": "revenueHigh",
        "estimated_revenue_avg": "revenueAvg",
        "estimated_sga_expense_low": "sgaExpenseLow",
        "estimated_sga_expense_high": "sgaExpenseHigh",
        "estimated_sga_expense_avg": "sgaExpenseAvg",
        "estimated_ebitda_low": "ebitdaLow",
        "estimated_ebitda_high": "ebitdaHigh",
        "estimated_ebitda_avg": "ebitdaAvg",
        "estimated_ebit_low": "ebitLow",
        "estimated_ebit_high": "ebitHigh",
        "estimated_ebit_avg": "ebitAvg",
        "estimated_net_income_low": "netIncomeLow",
        "estimated_net_income_high": "netIncomeHigh",
        "estimated_net_income_avg": "netIncomeAvg",
        "estimated_eps_low": "epsLow",
        "estimated_eps_high": "epsHigh",
        "estimated_eps_avg": "epsAvg",
        "number_analysts_estimated_revenue": "numAnalystsRevenue",
        "number_analysts_eps": "numAnalystsEps",
    }


class FMPAnalystEstimatesFetcher(
    Fetcher[
        FMPAnalystEstimatesQueryParams,
        list[FMPAnalystEstimatesData],
    ]
):
    """Transform the query, extract and transform the data from the FMP endpoints."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPAnalystEstimatesQueryParams:
        """Transform the query params."""
        return FMPAnalystEstimatesQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPAnalystEstimatesQueryParams,
        credentials: Optional[dict[str, str]],
        **kwargs: Any,
    ) -> list[dict]:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        import asyncio  # noqa
        import warnings
        from openbb_core.provider.utils.helpers import amake_request
        from openbb_fmp.utils.helpers import response_callback

        api_key = credentials.get("fmp_api_key") if credentials else ""

        symbols = query.symbol.split(",")  # type: ignore

        results: list[dict] = []

        async def get_one(symbol):
            """Get data for one symbol."""
            url = (
                "https://financialmodelingprep.com/stable/analyst-estimates?"
                + f"symbol={symbol}&period={query.period}"
                + f"&page={query.page if query.page else 0}&limit={query.limit if query.limit else 1000}"
                + f"&apikey={api_key}"
            )
            result = await amake_request(
                url, response_callback=response_callback, **kwargs
            )
            if not result or len(result) == 0:
                warnings.warn(f"Symbol Error: No data found for {symbol}")
            if result:
                results.extend(result)

        await asyncio.gather(*[get_one(symbol) for symbol in symbols])

        if not results:
            raise EmptyDataError("No data returned for the given symbols.")

        return sorted(results, key=lambda x: (x["date"], x["symbol"]), reverse=False)

    @staticmethod
    def transform_data(
        query: FMPAnalystEstimatesQueryParams, data: list[dict], **kwargs: Any
    ) -> list[FMPAnalystEstimatesData]:
        """Return the transformed data."""
        return [FMPAnalystEstimatesData.model_validate(d) for d in data]



"""FMP Historical Market Cap Model."""

# pylint: disable=unused-argument

from datetime import datetime
from typing import Any, Optional

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.historical_market_cap import (
    HistoricalMarketCapData,
    HistoricalMarketCapQueryParams,
)
from openbb_core.provider.utils.errors import EmptyDataError


class FmpHistoricalMarketCapQueryParams(HistoricalMarketCapQueryParams):
    """FMP Historical Market Cap Query.

    Source: https://site.financialmodelingprep.com/developer/docs#historical-market-cap

    """

    __json_schema_extra__ = {
        "symbol": {"multiple_items_allowed": True},
    }


class FmpHistoricalMarketCapData(HistoricalMarketCapData):
    """FMP Historical Market Cap Data."""

    __alias_dict__ = {
        "market_cap": "marketCap",
    }


class FmpHistoricalMarketCapFetcher(
    Fetcher[
        FmpHistoricalMarketCapQueryParams,
        list[FmpHistoricalMarketCapData],
    ]
):
    """FMP Historical Market Cap Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FmpHistoricalMarketCapQueryParams:
        """Transform the query params."""
        return FmpHistoricalMarketCapQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FmpHistoricalMarketCapQueryParams,
        credentials: Optional[dict[str, str]],
        **kwargs: Any,
    ) -> list:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        import warnings  # noqa
        from dateutil.relativedelta import relativedelta
        from openbb_fmp.utils.helpers import get_data_urls

        symbols = query.symbol.split(",")
        api_key = credentials.get("fmp_api_key") if credentials else ""
        results: list = []

        def generate_urls(symbol, start_date, end_date):
            """Generate URLs for each 5-year interval between start_date and end_date."""
            urls: list = []
            base_url = f"https://financialmodelingprep.com/stable/historical-market-capitalization?symbol={symbol}&limit=5000"
            base_url = base_url + "&from={}&to={}"

            while start_date <= end_date:
                next_date = start_date + relativedelta(months=60)
                url = base_url.format(
                    start_date.strftime("%Y-%m-%d"),
                    min(next_date, end_date).strftime("%Y-%m-%d"),
                )
                url = url + f"&apikey={api_key}"
                urls.append(url)
                start_date = next_date

            return urls

        for symbol in symbols:
            end_date = (
                query.end_date
                if query.end_date is not None
                else datetime.today().date()
            )

            urls = (
                generate_urls(symbol, query.start_date, end_date)
                if query.start_date and end_date
                else [
                    f"https://financialmodelingprep.com/stable/historical-market-capitalization?symbol={symbol}&limit=5000&apikey={api_key}"
                ]
            )
            data = await get_data_urls(urls, **kwargs)

            if not data:
                warnings.warn(f"No data was found for: {symbol}")
                continue

            results.extend(data)

        return results

    @staticmethod
    def transform_data(
        query: FmpHistoricalMarketCapQueryParams,
        data: list,
        **kwargs: Any,
    ) -> list[FmpHistoricalMarketCapData]:
        """Return the transformed data."""
        if not data:
            raise EmptyDataError("No data was returned for the given symbols.")
        symbols = query.symbol.split(",")
        return [
            FmpHistoricalMarketCapData.model_validate(d)
            for d in sorted(
                data,
                key=lambda x: (
                    x.get("date", len(symbols)),
                    -(x.get("marketCap", 0) or 0),
                ),
            )
        ]



"""FMP ETF Info Model."""

# pylint: disable=unused-argument

from datetime import datetime
from typing import Any, Optional, Union

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.etf_info import (
    EtfInfoData,
    EtfInfoQueryParams,
)
from openbb_core.provider.utils.errors import EmptyDataError
from pydantic import ConfigDict, Field, field_validator


class FMPEtfInfoQueryParams(EtfInfoQueryParams):
    """FMP ETF Info Query."""

    __json_schema_extra__ = {"symbol": {"multiple_items_allowed": True}}


class FMPEtfInfoData(EtfInfoData):
    """FMP ETF Info Data."""

    model_config = ConfigDict(extra="ignore")

    __alias_dict__ = {
        "issuer": "etfCompany",
        "cusip": "securityCusip",
        "aum": "assetsUnderManagement",
        "nav": "netAssetValue",
        "currency": "navCurrency",
        "volume_avg": "avgVolume",
        "updated": "updatedAt",
    }

    cusip: Optional[str] = Field(default=None, description="CUSIP of the ETF.")
    isin: Optional[str] = Field(default=None, description="ISIN of the ETF.")
    asset_class: Optional[str] = Field(
        default=None, description="Asset class of the ETF."
    )
    currency: Optional[str] = Field(
        default=None, description="Currency of the ETF's net asset value."
    )
    holdings_count: Optional[int] = Field(
        default=None, description="Number of holdings."
    )
    aum: Optional[float] = Field(
        default=None,
        description="Assets under management.",
        json_schema_extra={"x-unit_measurement": "currency"},
    )
    expense_ratio: Optional[float] = Field(
        default=None,
        description="The expense ratio, as a normalized percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    nav: Optional[float] = Field(
        default=None,
        description="Net asset value of the ETF.",
        json_schema_extra={"x-unit_measurement": "currency"},
    )
    volume_avg: Optional[Union[int, float]] = Field(
        default=None, description="Average daily trading volume."
    )
    updated: Optional[datetime] = Field(
        default=None, description="As of date for the latest data point."
    )

    @field_validator("expense_ratio", mode="before", check_fields=False)
    @classmethod
    def _normalize_percent(cls, v):
        """Normalize percent values."""
        return v / 100 if v else None


class FMPEtfInfoFetcher(
    Fetcher[
        FMPEtfInfoQueryParams,
        list[FMPEtfInfoData],
    ]
):
    """FMP ETF Info Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> FMPEtfInfoQueryParams:
        """Transform the query."""
        return FMPEtfInfoQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: FMPEtfInfoQueryParams,
        credentials: Optional[dict[str, str]],
        **kwargs: Any,
    ) -> list:
        """Return the raw data from the FMP endpoint."""
        # pylint: disable=import-outside-toplevel
        import asyncio  # noqa
        import warnings
        from openbb_fmp.utils.helpers import get_data_many

        api_key = credentials.get("fmp_api_key") if credentials else ""
        symbols = query.symbol.split(",")
        results: list = []

        async def get_one(symbol):
            """Get one symbol."""
            url = f"https://financialmodelingprep.com/stable/etf/info?symbol={symbol}&apikey={api_key}"
            response = await get_data_many(url, **kwargs)
            if not response:
                warnings.warn(f"No results found for {symbol}.")
            results.extend(response)

        await asyncio.gather(*[get_one(symbol) for symbol in symbols])

        return sorted(
            results,
            key=(lambda item: (symbols.index(item.get("symbol", len(symbols))))),
        )

    @staticmethod
    def transform_data(
        query: FMPEtfInfoQueryParams, data: list, **kwargs: Any
    ) -> list[FMPEtfInfoData]:
        """Return the transformed data."""
        if not data:
            raise EmptyDataError("No data was found for the given symbols.")
        return [FMPEtfInfoData.model_validate(d) for d in data]
